"""[EXPERIMENTAL] Tools for creating and managing menus."""
import warnings
warnings.warn(f"The {__name__!r} module and its' contents are experimental and may change without notice.")

import collections
from ._typing import (
    Any,
    Callable,
    Sequence,
    overload,
    override,
    Any,
    Item,
)
from . import _interface, api,items, color, style, events
from .api import Viewport


__all__ = [
    'close_context_menus',
    'close_context_menus_from',
    'hovered_context_menu',

    'ContextMenu',
    'ContextItem',
]




class _CtxMenuStack(collections.deque[tuple['ContextMenu', 'ContextItem | None']]):
    __slots__ = ('_lock',)

    __ctx_hregistry: items.mvHandlerRegistry

    # Each pair in the stack contains a context menu and the
    # menu item that spawned it (hovered). If the menu item
    # (2nd value) is None, the menu (1st value) is the origin
    # context menu and not a sub menu. The stack will always
    # properly hide menus as they are popped, but will not
    # display them when added.

    def __init__(self, *, maxlen: int | None = None) -> None:
        super().__init__(maxlen=maxlen)
        self._lock = api.mutex()

    def __cb_close_all(self, sender, mdown_info: tuple[int,  float]):
        '''Close all context menus when a mouse down event is
        captured while the cursor is away from them.'''
        duration = mdown_info[1]
        events.Py_DECREF(mdown_info)
        # ignore long-press/drag events
        if duration >= 0.05 or self.hovered_menu():
            return
        self.close_menus()  # disables hregistry

    def __get_ctx_hregistry(self):
        try:
            return self.__ctx_hregistry

        except AttributeError:
            hr_id = f'[{items.mvHandlerRegistry.__name__}] {__name__}'

            if api.Registry.alias_exists(hr_id):
                type(self).__ctx_hregistry = items.mvHandlerRegistry.new(hr_id)

            else:
                with items.mvHandlerRegistry.aliased(tag=hr_id, show=False) as type(self).__ctx_hregistry:
                    items.mvMouseDownHandler(callback=self.__cb_close_all)

        return self.__ctx_hregistry

    def _enable_ctx_hregistry(self):
        self.__get_ctx_hregistry().configure(show=True)

    def _disable_ctx_hregistry(self):
        self.__get_ctx_hregistry().configure(show=False)

    def hovered_menu(self):
        with self._lock:
            csr_x_pos, csr_y_pos = Viewport.mouse_pos_global()
            for ctx_branch in self:
                state = ctx_branch[0].state()
                x_pos, y_pos   = state['pos']        # type: ignore
                x_size, y_size = state['rect_size']  # type: ignore
                if (
                    (x_pos <= csr_x_pos <= x_pos + x_size)
                    and
                    (y_pos <= csr_y_pos <=  y_pos + y_size)
                ):
                    return ctx_branch
            return None

    def push_menu(self, menu: 'ContextMenu', menu_item: 'ContextItem | None' = None):
        """Push a context menu branch onto the stack."""
        with self._lock:
            self.appendleft((menu, menu_item))
            self._enable_ctx_hregistry()

    def pop_menu(self):
        """Pop the most-recent context menu branch off the stack."""
        with self._lock:
            menu = self.popleft()
            menu[0].configure(
                show=False,
                no_bring_to_front_on_focus=False,
            )
            return menu

    # TODO: trigger window `on_close` callbacks

    def close_menus_from(self, menu: 'ContextMenu', menu_item: 'ContextItem | None' = None):
        with self._lock:
            if not self:
                # `menu` is the origin and not a sub menu
                self._enable_ctx_hregistry()
                return self.appendleft((menu, None))

            # XXX: An `IndexError` here is a problem, but not the
            # stacks' problem.
            while True:
                sub_menu, owner_item = self.popleft()
                if sub_menu == menu:
                    self.appendleft((sub_menu, owner_item))
                    break
                elif owner_item != menu_item:
                    sub_menu.configure(
                        show=False,
                        no_bring_to_front_on_focus=False,
                    )

    def close_menus(self):
        with self._lock:
            self._disable_ctx_hregistry()
            for sub_menu, menu_item in self:
                sub_menu.configure(
                    show=False,
                    no_bring_to_front_on_focus=False,
                )
            self.clear()




_CTX_MENU_STACK = _CtxMenuStack()


def close_context_menus():
    """Dismiss all open context menus."""
    _CTX_MENU_STACK.close_menus()


def close_context_menus_from(menu: 'ContextMenu'):
    """Close all context menus and sub menus owned (directly
    or indirectly) by other menus.

    Args:
        * menu: Target context menu.


    If *menu* is not opened, all open context menus are
    closed.
    """
    with api.mutex():
        while _CTX_MENU_STACK:
            if _CTX_MENU_STACK[0][0] == menu:
                break
            _CTX_MENU_STACK.popleft()[0].configure(
                show=False,
                no_bring_to_front_on_focus=False,
            )


def hovered_context_menu():
    """Return the open context menu or sub menu under the
    cursor.

    The result of this function is independent of the menu's
    "hovered" state and may return a menu even when that menu's
    "hovered" state is False. This is because item children
    can "own" state over their parents. Instead, this function
    compares the explicit positions of the cursor and all open
    menus.
    """
    menu_branch = _CTX_MENU_STACK.hovered_menu()
    if menu_branch:
        return menu_branch[0]








def _get_default_deps_alias(cls: type, itp: type, component_name: str) -> str:
    "Returns the alias of a class' dependency item."
    return f'[{itp.__name__}] {cls.__name__}.{component_name}'.replace('..', '.')


def _get_default_theme_items(cls: type, component_name: str):
    "Returns a class' dependency theme item and components."
    theme_alias    = _get_default_deps_alias(cls, items.mvTheme, component_name)
    tc_color_alias = _get_default_deps_alias(
        cls, items.mvThemeComponent, f'{component_name}.color'
    ).replace('..', '.')
    tc_style_alias = _get_default_deps_alias(
        cls, items.mvThemeComponent, f'{component_name}.style'
    ).replace('..', '.')

    if api.Registry.alias_exists(theme_alias):
        theme = items.mvTheme.new(theme_alias)

        if api.Registry.alias_exists(tc_color_alias):
            tc_color = items.mvThemeComponent.new(tc_color_alias)
        else:
            tc_color = items.mvThemeComponent.aliased(tag=tc_color_alias, parent=theme)

        if api.Registry.alias_exists(tc_style_alias):
            tc_style = items.mvThemeComponent.new(tc_style_alias)
        else:
            tc_style = items.mvThemeComponent.aliased(tag=tc_style_alias, parent=theme)

    else:
        with items.mvTheme.aliased(tag=theme_alias) as theme:
            tc_color = items.mvThemeComponent.aliased(tag=tc_color_alias)
            tc_style = items.mvThemeComponent.aliased(tag=tc_style_alias)

    return theme, tc_color, tc_style


def _ensure_theme_elements(theme_component: items.mvThemeComponent, *elems: tuple[Callable, tuple[float, ...]]):
    'Creates missing theme element items of dependency items.'
    tc_alias = theme_component.alias
    with theme_component:
        for fn, elem_v in elems:
            alias = f'{tc_alias}.{fn.__name__}'
            if not api.Registry.alias_exists(alias):
                element = fn(*elem_v)
                element.alias = alias


def _fix_menu_position(menu_x_size: int, menu_y_size: int, x_pos: int, y_pos: int):
    "Adjust a menu's new position depending on the menu's size."
    # XXX: The only difference between this positioning behavior
    # and `MenuItem._cb_hovered`s behavior is that padding isn't
    # a factor here.
    _vp_cfg = Viewport.configuration()
    vp_wt = _vp_cfg['client_width']  # type: ignore
    vp_ht = _vp_cfg['client_height'] # type: ignore
    if menu_x_size > vp_wt - x_pos:
        x_pos -= menu_x_size
    if menu_y_size > vp_ht - y_pos:
        y_pos = max(0, y_pos - menu_y_size)
    return x_pos, y_pos



class ContextItem(_interface.SupportsCallback, items.mvGroup):
    """A highly-customizable implementation of `mvMenuItem` for
    vertical context menus.

    > NOTE: `ContextItem`s are **compound items**; composed of several
    other items. They may break when configured using Dear PyGui's API.
    """
    # Expect some 'excessive' block comments here. This
    # implementation is confusing at a glance, but addresses many
    # issues other implementations may have;
    #   - Dependency data is not stored onto the interface, unless
    #   it can be "handled" when missing.
    #
    #   - It acts like any other built-in interface/item. Users
    #   should not need to walk on eggshells when using them (i.e.
    #   minimal 'gacha's).
    #
    #   - Most customization (color, padding, size, etc.) is made by
    #   tweaking theme element values. The default look is inherited
    #   by the parent menu, and adjustments can be made per-item by
    #   binding it to an appropriate theme. Best part is that the
    #   menu theme can just be copied (`menu.theme.__copy__()`),
    #   set onto the menu item, and adjusted from there; zero element
    #   conflicts.

    __c_area_theme: items.mvTheme
    __spacer_theme: items.mvTheme

    @overload
    def __init__(  # type: ignore
        self,
        label: str,
        *,
        callback          : Callable | None = ...,
        user_data         : Any = ...,
        use_internal_label: bool = ...,
        width             : int = ...,
        height            : int = ...,
        indent            : int = ...,
        before            : Item = ...,
        payload_type      : str = ...,
        drag_callback     : Callable | None = ...,
        drop_callback     : Callable | None = ...,
        show              : bool = ...,
        pos               : tuple[int, int] | list[int] = ...,
        filter_key        : str = ...,
        delay_search      : bool = ...,
        tracked           : bool = ...,
        track_offset      : float = ...,
        **kwargs,
    ): ...
    def __init__(self, label: str, *, callback: Callable | None = None, **kwargs) -> None:
        super().__init__(**kwargs)

        with events.root_ihandler_registry(self):
            items.ResizeHandler(callback=self._cb_resized)

        if not self.handlers:
            with items.mvItemHandlerRegistry() as self.handlers:
                items.mvHoverHandler(callback=self._cb_hovered)

        with self:
            # The button below is the menu item's "click area".
            # The label layout is within the group further down.
            # "Why a button? Why not just an in-table selectable
            # w/`span_columns=True`?"
            #   - Buttons support more customizations w/theme elements.
            #   Do you like rounded buttons? I like rounded buttons.
            #
            #   - Customization is MUCH easier w/themes vs different
            #   config. options spread thin over several items.
            #
            # In addition, the button's user_data is used internally
            # over the "face value" group item to that it's free for
            # users to use.
            button = items.mvButton(
                width=-1,
                height=1,
                callback=self._cb_clicked,
            )
            try:
                button.theme = self.__c_area_theme
            except AttributeError:
                theme, tc_color, tc_style = _get_default_theme_items(type(self), 'click_area')
                button.theme = self.__class__.__c_area_theme = theme
                _ensure_theme_elements(
                    tc_color,
                    (color.button, (0, 0, 0, 0)),
                )

            # Two reasons for this inner group. First, the click area
            # needs to scale with the table, which is a problem since
            # tables cannot be explicitly positioned and do not support
            # states/handlers. Groups support many states/handlers,
            # with a rect calculated from its' contents.
            # The second reason is that last bit -- We ONLY want the
            # table size, so the click area and table cannot be
            # parented by the same group.
            # BUG: Updates to most relevant style elements will trigger
            # a resize event on the parenting menu. However, some do not,
            # and others only do so indirectly (if, for example, the
            # parent is auto-sized). If the parent is auto-sized, the
            # vertical component of `frame_padding` and `cell_padding`
            # won't trigger one if the result causes the parent to shrink.
            with items.mvGroup(label="[1] layout"):
                with items.mvTable(
                    header_row=False,
                    no_host_extendX=True,
                    no_keep_columns_visible=True,
                    context_menu_in_body=False,
                    sortable=False,
                    resizable=False,
                    reorderable=False,
                    scrollY=False,
                    scrollX=False,
                    no_clip=True,
                ) as table:
                    with items.mvTableRow():
                        items.mvTableColumn(label='[0] padding', width_fixed=True, parent=table, no_clip=True, no_reorder=True, no_sort=True)
                        items.mvSpacer()

                        # menu name
                        items.mvTableColumn(label='[1] label (name)', width_fixed=True, parent=table, no_clip=True, no_reorder=True, no_sort=True)
                        name_lbl = items.mvText()

                        # The "spacer" is necessary for the group to calculate its
                        # size properly. An empty, stretchy column is not enough; it
                        # needs content to consume the space.
                        items.mvTableColumn(label='[2] spacing', width_stretch=True, parent=table, no_reorder=True, no_sort=True, indent_disable=True, indent_enable=False)
                        spacer = items.mvButton(width=-1, height=1)
                        try:
                            spacer.theme = self.__spacer_theme
                        except AttributeError:
                            theme, tc_color, tc_style = _get_default_theme_items(type(self), 'spacer')
                            spacer.theme = self.__class__.__spacer_theme = theme
                            _ensure_theme_elements(
                                tc_color,
                                (color.button, (0, 0, 0, 0)),
                                (color.button_active, (0, 0, 0, 0)),
                                (color.button_hovered, (0, 0, 0, 0)),
                            )

                        # menu shortcut
                        items.mvTableColumn(label='[3] label (shortcut)', width_fixed=True, parent=table, no_reorder=True, no_sort=True)
                        shortcut_lbl = items.mvText()

                        items.mvTableColumn(label='[4] padding', width_fixed=True, parent=table, no_clip=True, no_reorder=True, no_sort=True)
                        items.mvSpacer()

        self._internal_data = (name_lbl, shortcut_lbl, None, None)
        self.configure(label=label, callback=callback)

    @property
    def root_parent(self):
        return ContextMenu.new(api.Item.root_parent(self))

    @property
    def _internal_data(self) -> tuple[items.mvText, items.mvText, 'ContextMenu | None', Callable | None]:
        # store in click_area (button)
        return api.Item.configuration(self.children(1)[0])['user_data']
    @_internal_data.setter
    def _internal_data(self, value: tuple[items.mvText, items.mvText, 'ContextMenu | None', Callable | None]):
        # XXX: The submenu is set by the menus themselves and
        # not this item.
        api.Item.configure(self.children(1)[0], user_data=value)

    def _set_submenu(self, menu: 'ContextMenu | None', ):
        name_item, shtcut_item, _, callback = self._internal_data
        self._internal_data = name_item, shtcut_item, menu, callback

    def _cb_resized(self, handler: Item = 0):
        'Triggers when the parenting root item is resized.'
        try:
            button, layout = self.children(1)
        except SystemError:
            if handler and not api.Registry.item_exists(self):
                return api.Item.destroy(handler)
            raise
        layout_state = api.Item.state(layout)
        api.Item.configure(
            button,
            pos=layout_state['pos'],
            height=layout_state['rect_size'][1],    # type: ignore
        )

    def _cb_hovered(self, *args):
        'Triggers when the click area is hovered.'
        # TODO: show cascade menu on hover (popup)
        smenu = self._internal_data[2]
        pmenu = self.root_parent

        _CTX_MENU_STACK.close_menus_from(pmenu, self)
        if smenu:
            _CTX_MENU_STACK.push_menu(smenu, self)
            if not smenu.is_visible:
                # HACK: imgui won't calculate an item's size if it hasn't
                # been rendered yet. Show it off-screen and wait a frame.
                # BUG: This is enough to force imgui to calculate the
                # the menu's inner padding, but not enough to actually
                # size the menu's content. It risks getting clipped off-
                # screen when displaying for the first time.
                smenu_x_size, smenu_y_size = smenu.state()['rect_size']  # type: ignore
                if not (smenu_x_size or smenu_y_size):
                    smenu.configure(show=True, pos=(100 + Viewport.client_width, 0))
                    api.Runtime.split_frame()
                    smenu_x_size, smenu_y_size = smenu.state()['rect_size']  # type: ignore

                _vp_cfg = Viewport.configuration()
                vp_wt = _vp_cfg['client_width']  # type: ignore
                vp_ht = _vp_cfg['client_height'] # type: ignore

                _self_state = self.state()
                self_x_min, self_y_min = _self_state['rect_min']  # type: ignore
                self_x_max, self_y_max = _self_state['rect_max']  # type: ignore

                # right/left edge of sub menu aligns with x_max/x_min of this
                # item, accounting for 50% of the parent menu's inner padding
                _wndw_padx_offset = int(
                    (pmenu.state()['rect_size'][0] - _self_state['rect_size'][0]) * 0.25  # type: ignore
                )
                if smenu_x_size < vp_wt - self_x_max:
                    new_x_pos = self_x_max + _wndw_padx_offset
                else:
                    new_x_pos = self_x_min - smenu_x_size - _wndw_padx_offset

                if smenu_y_size < vp_ht - self_y_min:
                    # y_min of the sub menu('s first item) ~aligned with
                    # y_min of this menu item
                    try:
                        new_y_pos = self_y_min - api.Item.state(  # type: ignore
                            smenu.information()['children'][1][0]
                        )['pos'][1]
                    except IndexError:
                        new_y_pos = self_y_min
                else:
                    new_y_pos = max(0, vp_ht - smenu_y_size)

                smenu.configure(show=True, pos=(new_x_pos, new_y_pos), no_bring_to_front_on_focus=False)
                pmenu.configure(no_bring_to_front_on_focus=True)

    def _cb_clicked(self):
        'Triggers when the click area is clicked.'
        callback = self._internal_data[-1]
        if callback:
            _CTX_MENU_STACK.close_menus()
            # Invokes the callback set by the user as if called by
            # DPG. This is not ideal (i.e. variadics count as 1 arg),
            # but it is the fastest (and perhaps the most expected)
            # way.
            callback(
                *(self, None, self.user_data)[:callback.__code__.co_argcount]
            )

    @override
    def configure(self, **kwargs):
        name_item, shtcut_item, sub_menu, callback = self._internal_data

        if 'label' in kwargs:
            name = kwargs['label']
            if '##' in name:
                name, shortcut = name.split('##')
            else:
                shortcut = ''

            name_item.value   = name
            shtcut_item.value = shortcut
            self._cb_resized()

        if 'callback' in kwargs:
            self._internal_data = (
                name_item,
                shtcut_item,
                sub_menu,
                kwargs.pop('callback')  # `mvGroup` does not support `callback`
            )

        super().configure(**kwargs)


class ContextMenu(items.mvWindowAppItem):
    """An extension of `mvWindowAppItem` used for creating
    and managing stylized context menus.

    The main appeal of using `ContextMenu`s over `mvMenu`s are
    their styling options. Built-in items like `mvMenu` and
    `mvMenuItem` have very rigid layouts and minimal customization.
    On the other hand, `ContextMenu` and `ContextItem` --
    `ContextMenu`s child of choice -- reflect a *lot* of theme
    elements. They're influenced by theme elements to such an
    extent that the `ContextMenu` class has its' own dedicated
    global theme, and exposes elements such as `frame_rounding`
    and `item_spacing` as instance properties. Additionally, users
    have a bit more control here than they would using built-in
    menu; sub-menus can be thrown as their own root context menu,
    and the hovered/in-focus menu can be easily obtained using the
    module-level `hovered_context_menu` function. And while they
    do not perfectly emulate the of built-in menus and menu items,
    their behavior is very close.


    Typically, a "context menu" is a menu that becomes visible
    on input (right-click, etc) or state change. `ContextMenu`
    instances are intended (but not limited) to replace the role
    of `mvMenu` items for that purpose. Like windows, they cannot be
    parented. However, context-manager usage mimics the tree-like
    nesting behavor of menu items when used within the statement
    body of other `ContextMenu` context managers:
        >>> import dearpypixl as dpx
        >>> from dearpypixl.menus import *
        >>>
        >>>
        >>> with ContextMenu("RootMenu") as ctx_menu:
        ...     ContextItem("Item1")
        ...     ContextItem("Item2", callback=lambda: print('clicked Item2'))
        ...
        ...     with ContextMenu("SubMenu1") as submenu_1:
        ...         dpx.ColorPicker([225, 100, 100, 255])
        ...         ContextItem("Item3")

    The menus are hidden by default. The menu's `show` configuration
    can still be updated to display/hide it like normal. The
    `.to_position` and `.to_cursor` methods are other convenient
    ways to display them on screen; taking into account the menu's
    content so that it is not clipped. When used for their intended
    purpose, opt to use the methods over the `show` configuration
    to ensure that the menu will hide when non-context menus are
    brought into focus or vice-versa.


    As with most other interfaces, `ContextMenu` instances are not
    required to create the items they manage, allowing for normal
    window items to be selectively repurposed as context menus.
    However, updates to the following configuration options should
    be avoided while the item is managed through the interface or
    `menu` module framework:
        - `no_focus_on_appearing`
        - `no_bring_to_front_on_focus`
        - `no_collapse`
        - `modal`
        - `popup`
    """
    __theme    : items.mvTheme

    mitem_factory: type[ContextItem] = ContextItem
    cascade_arrow: str               = '\u009b'

    label: str

    @overload
    def __init__(  # type: ignore
        self,
        label                      : str = '',
        *,
        width                      : int  = -1,
        height                     : int  = -1,
        show                       : bool = False,
        autosize                   : bool = True,
        no_resize                  : bool = True,
        no_title_bar               : bool = True,
        no_move                    : bool = True,
        user_data                  : Any = ...,
        use_internal_label         : bool = ...,
        tag                        : Item = ...,
        pos                        : tuple[int, int] | list[int] = ...,
        delay_search               : bool = ...,
        min_size                   : Sequence[int] = ...,
        max_size                   : Sequence[int] = ...,
        menubar                    : bool = ...,
        no_scrollbar               : bool = ...,
        horizontal_scrollbar       : bool = ...,
        no_close                   : bool = ...,
        indent                     : int = ...,
        no_background              : bool = ...,
        no_open_over_existing_popup: bool = ...,
        no_scroll_with_mouse       : bool = ...,
        on_close                   : Callable | None = ...,
        **kwargs,
    ): ...
    def __init__(
        self,
        label                     : str = '',
        *,
        width                     : int = -1,
        height                    : int = -1,
        show                      : bool = False,
        autosize                  : bool = True,
        no_resize                 : bool = True,
        no_title_bar              : bool = True,
        no_move                   : bool = True,
        min_size                  : tuple[int, int] | Sequence[int] = (5, 5),
        # ignore these kwds from caller
        no_focus_on_appearing     : Any = ...,
        no_bring_to_front_on_focus: Any = ...,
        no_collapse               : Any = ...,
        modal                     : Any = ...,
        popup                     : Any = ...,
        **kwargs,
    ) -> None:
        super().__init__(
            no_focus_on_appearing=False,
            no_bring_to_front_on_focus=False,
            no_collapse=True,
            modal=False,
            popup=False,
        )

        if not self.handlers:
            self.handlers = items.mvItemHandlerRegistry()

        if not self.theme:
            try:
                self.theme = self.__theme
            except AttributeError:
                theme, tc_color, tc_style = _get_default_theme_items(type(self), '')
                self.theme = self.__class__.__theme = theme
                _ensure_theme_elements(
                    tc_color,
                    (color.button, (0, 0, 0, 0)),
                    (color.button_active, (29, 151, 236, 103)),
                    (color.button_hovered, (29, 151, 236, 103)),
                )
                _ensure_theme_elements(
                    tc_style,
                    (style.item_spacing, (8, 10)),
                    (style.window_padding, (24, 8)),
                    (style.popup_rounding, (0, -1)),
                    (style.frame_rounding, (0, -1)),
                )

        self.configure(
            label=label,
            width=width,
            height=height,
            show=show,
            autosize=autosize,
            no_resize=no_resize,
            no_title_bar=no_title_bar,
            no_move=no_move,
            min_size=min_size,
            **kwargs
        )

    def _get_theme_element(self, component_name: str, element_name: str) -> Any:
        theme = self.theme
        if not theme or theme == self.__theme:
            alias = _get_default_deps_alias(
                self.__class__,
                items.ThemeComponent,
                component_name
            )
            return items.interface(f'{alias}.{element_name}')  # type: ignore
        for c in theme.children(1)[1]:  # type: ignore
            for e in api.Item.children(c, 1):
                if api.ThemeElement.identify(e) == element_name:
                    return items.interface(e)  # type: ignore
        raise ValueError(f'assigned theme has no element item {element_name!r}.')

    @property
    def button(self) -> items.mvThemeColor:
        return self._get_theme_element('color', 'button')

    @property
    def button_active(self) -> items.mvThemeColor:
        return self._get_theme_element('color', 'button_active')

    @property
    def button_hovered(self) -> items.mvThemeColor:
        return self._get_theme_element('color', 'button_hovered')

    @property
    def item_spacing(self) -> items.mvThemeStyle:
        return self._get_theme_element('style', 'item_spacing')

    @property
    def window_padding(self) -> items.mvThemeStyle:
        return self._get_theme_element('style', 'window_padding')

    @property
    def popup_rounding(self) -> items.mvThemeStyle:
        return self._get_theme_element('style', 'popup_rounding')

    @property
    def frame_rounding(self) -> items.mvThemeStyle:
        return self._get_theme_element('style', 'frame_rounding')

    # NOTE:'ctx' in this case refers to both the type of menu AND
    # language, as it's managed through the item's `with` statement
    # hooks. Completely independent and in no way related to
    # `_CTX_MENU_STACK`.
    __ctx_menu_stack: collections.deque[items.mvWindowAppItem] = collections.deque()

    @override
    def push_stack(self):
        stack = self.__class__.__ctx_menu_stack
        if not stack or stack[0] != self:
            stack.appendleft(self)
        return super().push_stack()

    @override
    def pop_stack(self):
        super().pop_stack()
        stack = self.__class__.__ctx_menu_stack
        if stack and stack[0] == self:
            stack.popleft()
            if stack and stack[0] != self:
                mitem = self.mitem_factory(label=self.label, parent=stack[0])
                mitem._set_submenu(self)
                self.configure()

    @overload
    def configure(self, *, label: str = ..., width: int = ..., height: int = ..., show: bool = ..., autosize: bool = ..., no_resize: bool = ..., no_title_bar: bool = ..., no_move: bool = ..., user_data: Any = ..., use_internal_label: bool = ..., tag: Item = ..., pos: tuple[int, int]|list[int] = ..., delay_search: bool = ..., min_size: Sequence[int] = ..., max_size: Sequence[int] = ..., menubar: bool = ..., no_scrollbar: bool = ..., horizontal_scrollbar: bool = ..., no_close: bool = ..., indent: int = ..., no_background: bool = ..., no_open_over_existing_popup: bool = ..., no_scroll_with_mouse: bool = ..., on_close: Callable|None = ..., **kwargs) -> None: ...  # type: ignore
    @override
    def configure(self, **kwargs):
        if 'label' in kwargs:
            label = kwargs['label']
            if not '##' in label:
                kwargs['label'] = f'{label}##{self.cascade_arrow}'

        super().configure(**kwargs)

    def to_position(self, x_pos: int, y_pos: int, *, content_aware: bool = True):
        """Close all context menus, then show this menu at the
        the given position. The menu will be brought into the
        foreground as the active window.

        Args:
            * x_pos: The menu's new horizontal position.

            * y_pos: The menu's new vertical position.

            * content_aware: If True (default), try to position the
            menu in a way that would avoid clipping its' content. If
            False, always positions the menu's top-left corner at
            the given coordinates.


        This is a no-op when the cursor is positioned over an open
        context menu.
        """
        if not _CTX_MENU_STACK.hovered_menu():
            _CTX_MENU_STACK.close_menus()
            self.configure(
                pos=(x_pos, y_pos) if not content_aware else _fix_menu_position(
                    *self.state()['rect_size'],
                    x_pos, # type: ignore
                    y_pos, # type: ignore
                ),
                show=True,
                no_bring_to_front_on_focus=False,
            )
            _CTX_MENU_STACK.push_menu(self)

    def to_cursor(self):
        """Close all context menus, then show this menu at the
        cursor's position. The menu will be brought into the
        foreground as the active window.

        This is a no-op when the cursor is positioned over an open
        context menu.
        """
        if not _CTX_MENU_STACK.hovered_menu():
            _CTX_MENU_STACK.close_menus()
            self.configure(
                pos=_fix_menu_position(
                    # BUG: Just like with `MenuItem`s on-hover behavior, this
                    # menu's size will be 0 if it has never been rendered. The
                    # menu's content is at risk of getting clipped for its'
                    # first viewing.
                    *self.state()['rect_size'],
                    *Viewport.mouse_pos_global()
                ),
                show=True,
                no_bring_to_front_on_focus=False,
            )
            _CTX_MENU_STACK.push_menu(self)



import re
import ast
import sys
import types
import inspect
import itertools  # type: ignore
import dataclasses
from inspect import Parameter
from types import ModuleType
from typing import (
    MutableMapping,
    TypeVar,
    Iterable,
    Sequence,
    Mapping,
    Any,
    Callable,
    Self,
    get_overloads,
    TYPE_CHECKING,
)
if TYPE_CHECKING:
    from ._interface import AppItemType



_T = TypeVar('_T')


# [ STRING FORMATTING ]

class _IndentedStrType(type):
    __slots__ = ()

    def indent(self) -> None:
        self.level += 1

    def dedent(self) -> None:
        self.level -= 1
        if self.level < 0:
            self.level = 0

    def __enter__(self):
        self.indent()

    def __exit__(self, *args):
        self.dedent()

class indent(str, metaclass=_IndentedStrType):
    __slots__ = ()

    __indent: str = ' ' * 4

    level: int = 0

    def __new__(cls, s: str, level: int | None = None) -> str:
        return cls.__indent * (level or cls.level) + s or ''

    @classmethod
    def lines(cls, s: str, level: int, *, sep: str = '\n') -> str:
        indent = cls.__indent * (level or cls.level)
        return indent + indent.join(s.split(sep))


# >>> re.findall('vector: collections.abc.Sequence[int | typing.Any]')
# ['collections.abc.S', 'typing.A']
_RE_PTRN_TPLOOKUP = re.compile(r'(?:(?::\s?)|(?:\[)|(?:\s*\|\s*)\s?)((?:(?:\w*\.)(?:\w))+)')

def _rm_source_anno_namespaces(s: str) -> str:
    """Removes dotted namespace lookup prefixes from signatures and
    annotations of Python source code.

        >>> clean_source_annotations('vector: collections.abc.Sequence[int | typing.Any]')
        'vector: Sequence[int | Any]'

    """
    # This should not remove namespace lookup strings elseware, but
    # needs more testing. Unused for now.
    for hit in set(re.findall(_RE_PTRN_TPLOOKUP, s)):
        try:
            # the pattern isn't perfect -- matches also include
            # the char after the last dot
            s = s.replace(hit[:-1], "")
        except IndexError:
            pass
    return s


_RE_PTRN_NSPFX = r'((?:(?:\w*\.))+)'

def trim_lookup_trail(s: str) -> str:
    """Removes dotted namespace lookup prefixes in a
    Python source code string."""
    # Unlike the *other* one, this one is rather...indescriminate
    # when it comes to matches.
    for hit in set(re.findall(_RE_PTRN_NSPFX, s)):
        s = s.replace(hit, "")
    return s


def uneval(tp: type | str | None) -> str:
    """Return a string that would evaluate to the passed type or
    annotation of that type when compiled as source code."""
    return inspect.formatannotation(tp)


def object_annotation(o: Any) -> str:
    """Return a string that would evaluate to the passed type or
    annotation of that type when compiled as source code if
    available in the would-be namespace."""
    return trim_lookup_trail(uneval(o))




# [ PARSERS / WRITERS ]

def _get_module(module: str | types.ModuleType) -> types.ModuleType:
    if isinstance(module, str):
        try:
            return sys.modules[module]
        except KeyError:
            raise ValueError(
                f"{module!r} is not the name of an imported module."
            ) from None
    assert isinstance(module, types.ModuleType)
    return module


@dataclasses.dataclass(init=False)
class Exported(MutableMapping):
    """A mapping containing names and objects to export to stub files.

    Flagging objects for export:
        >>> exported = Exported(__name__)  # `ModuleType` object also works
        ...
        >>> @exported  # uses `PublicClass.__qualname__` as the key
        >>> class PublicClass: ...
        ...
        >>> @exported  # uses `public_func.__name__` as the key
        >>> def public_func(): ...
        ...
        >>> # passing an explicit name
        >>> exported(public_func, "alias_of_public_func")

    Getting exports from a module:
        >>> # the module must have been imported prior
        >>> exports = Exported.fetch(__name__)  # `ModuleType` object also works

    """

    module: str | types.ModuleType

    __slots__ = ("_module", "_exported", "aliased")

    EXPORTED = '__exported__'
    ALIASED  = '__aliased__'

    def __init__(self, module: str | types.ModuleType):
        self._exported = {}
        self.aliased   = {}
        self._module   = _get_module(module)
        setattr(self._module, self.EXPORTED, self)

    def __getitem__(self, key: str):
        return self._exported[key]

    def __setitem__(self, key: str, value: Any):
        self._exported[key] = value

    def __delitem__(self, key: str):
        del self._exported[key]

    def __iter__(self):
        return iter(self._exported)

    def __len__(self):
        return len(self._exported)

    def export(self, o: _T | None = None, /, name: str = '') -> _T:
        def capture_object(o: _T) -> _T:
            nonlocal name
            if not name:  # type: ignore
                try:
                    name = o.__qualname__  # type: ignore
                except AttributeError:
                    name = o.__name__   # type: ignore
            self._exported[name] = o  # type: ignore
            return o
        if o is None:
            return capture_object  # type: ignore
        return capture_object(o)

    __call__ = export

    def export_alias(self, name: str, o: _T | None = None) -> _T:
        def capture_object(o: _T) -> _T:
            self.aliased[name] = o  # type: ignore
            return o
        if o is None:
            return capture_object  # type: ignore
        return capture_object(o)

    @classmethod
    def fetch(cls, module: types.ModuleType) -> Self:
        """Return a mapping of a module's exports."""
        module = _get_module(module)
        return getattr(module, cls.EXPORTED, None) or cls(module)





@dataclasses.dataclass(slots=True)
class Imported(Sequence):
    module  : str | ModuleType
    sorted  : bool       = True
    absolute: bool       = False
    relative: bool       = True

    _name_map: dict[Any, str] = dataclasses.field(init=False, repr=False)
    _imports : list[Any]      = dataclasses.field(init=False, repr=False)
    _exports : list[Any]      = dataclasses.field(init=False, repr=False)

    def __post_init__(self):
        self.module    = _get_module(self.module)
        self._imports  = []
        self._exports  = []   # "exported" imports need to be followed up with an alias "obj as obj"
        self._name_map = {}
        self._build_name_map()

    def _build_name_map(self):
        self._name_map.clear()
        for k, v in self.module.__dict__.items():
            try:
                if v not in self._name_map:
                    self._name_map[v] = k
            except TypeError:
                continue

    def __getitem__(self, key: int) -> Any:
        return NotImplemented

    def __len__(self):
        return len(self._imports) + len(self._exports)

    def __iter__(self):
        yield from itertools.chain(self._imports, self._exports)

    def __str__(self) -> str:
        stmt = []
        src  = self.module.__name__  # type: ignore
        if self.absolute and self.relative:
            if not self._imports:
                return ''
            stmt.append(f"from {src} import (")
        elif self.relative:
            if not self._imports and not self._exports:
                return ''
            stmt.append(f'from .{src.split(".", maxsplit=1)[-1]} import (')
        else:
            return f'import {src}'

        imports = []
        for o in self._imports:
            imports.append(f"    {self._name_map[o]},")
        for o in self._exports:
            try:
                o_name = self._name_map[o]
            except KeyError:
                if isinstance(o, str) and hasattr(self.module, o):
                    o_name = o
                else:
                    raise
            imports.append(f"    {o_name} as {o_name},")

        if self.sorted:
            imports.sort()

        stmt.extend(imports)
        stmt.append(")")
        return '\n'.join(stmt)

    @property
    def imports(self) -> tuple:
        return tuple(self._imports)

    @property
    def exports(self) -> tuple:
        return tuple(self._exports)

    def copy(self) -> Self:
        copy = type(self)(self.module, self.sorted, self.absolute, self.relative)
        copy._imports = self._imports.copy()
        return copy

    def append(self, value: Any, *, export: bool = False):
        if export:
            self._exports.append(value)
            if value in self._imports:
                self._imports.remove(value)
        else:
            self._imports.append(value)
            if value in self._exports:
                self._exports.remove(value)

    def extend(self, iterable: Iterable[Any], *, export: bool = False):
        for v in iterable:
            self.append(v, export=export)




def validate_source(s: str) -> str:
    """Validate a Python source code string.

    Returns the string unchanged if the source is valid. Throws
    various errors otherwise.
    """
    ast.parse(s)
    return s


def to_source_classdef(tp: type) -> str:
    """Return a source code string containing the preliminary
    class definition of a type.
    """
    src = [
        f"class {tp.__qualname__}({', '.join(b.__qualname__ for b in tp.__bases__)}):"
    ]
    with indent:
        if tp.__doc__:
            src.append(indent('"""') + tp.__doc__  + '\n    """')

        if '__slots__' in tp.__dict__:
            slots = tp.__slots__  # type: ignore
            if slots:
                src.append(indent('__slots__: tuple[str, ...] = ('))
                with indent:
                    mbrs = ''.join(
                        ",\n".join(indent(repr(m)) for m in slots)
                    )
                    src.append(mbrs)
                src.append(indent(")"))
            else:
                src.append(indent('__slots__: tuple[str, ...] = ()'))
    return '\n'.join(src)


def to_source_parameters(parameters: Mapping[str, Parameter]) -> str:
    """Return a source code string containing the argument
    signature of a callable.
    """
    src = []

    mk_pos_only = False
    mk_kwd_only = False
    for p in parameters.values():
        if p.annotation is p.empty:
            src_string = p.name
        else:
            src_string = f"{p.name}: {trim_lookup_trail(uneval(p.annotation))}"
        if p.default is not p.empty:
            src_string = f"{src_string} = ..."

        kind = p.kind

        if kind == p.POSITIONAL_ONLY:
            mk_pos_only = True
            src.append(src_string)
            continue
        elif mk_pos_only:
            mk_pos_only = False
            src_string = f"/, {src_string}"

        if kind == p.VAR_POSITIONAL:
            # Arguments following VAR_POSITIONAL are made KEYWORD_ONLY
            # by way of Python syntax. In this case, however, the lone
            # "*" marker is not needed.
            mk_kwd_only = True
        elif kind == p.KEYWORD_ONLY and not mk_kwd_only:
            src_string  = f"*, {src_string}"
            mk_kwd_only = True
        elif kind == p.VAR_KEYWORD:
            src_string  = f"**{src_string}"

        src.append(src_string)

    return ', '.join(src).replace("*, /,", "/, *,")


def to_source_function(fn: Callable):
    pyi = []

    overloads = get_overloads(fn)
    if overloads:
        for ovrld in overloads:
            sig = inspect.signature(ovrld)
            params = to_source_parameters(sig.parameters)
            return_tp = object_annotation(sig.return_annotation)
            pyi.extend((
                f"@overload",
                f"def {fn.__name__}({params}) -> {return_tp}: ...",
            ))

    else:
        sig = inspect.signature(fn)
        params = to_source_parameters(sig.parameters)
        return_tp = object_annotation(sig.return_annotation)
        pyi.append(f"def {fn.__name__}({params}) -> {return_tp}: ...")

    return '\n'.join(pyi)



"""Contains functions that return preset themes.

A preset is created only if it doesn't exist. Otherwise,
the existing preset(s) are returned. Each preset is
assigned an alias that remains constant throughout
sessions. The collection of available Dear PyPixl presets
can be obtained by calling the `load_presets` function.

Presets can be cloned by calling their `.__copy__` method.
"""
import functools
import contextlib
from . import _typing, api
from . import color, style
from .items import (
    mvTheme,
    mvThemeComponent,
)




def _onetime_setup():
    if not api.Application.state()['ok']:
        api.Application.setup()
    global _onetime_setup
    _onetime_setup = lambda: None


_preset_fns: tuple[_typing.Callable[[], mvTheme], ...] = ()

def _dearpypixl_preset(preset_name: str):
    def capture_fn(fn: _typing.Callable[[], mvTheme]):
        @functools.wraps(fn)
        def get_theme_preset():
            _onetime_setup()
            if api.Registry.alias_exists(preset_name):
                return mvTheme.new(preset_name)
            return fn()
        global _preset_fns
        _preset_fns += (fn,)
        return get_theme_preset
    preset_name = f'[{mvTheme.__name__}] {preset_name}'
    return capture_fn


@contextlib.contextmanager
def _theme_preset(name: str):
    with mvTheme.aliased(tag=f'[mvTheme] {name}', label=f'[mvTheme] {name}') as theme:
        color_c = mvThemeComponent.aliased(
            tag=f'[mvThemeComponent] color ({name})',
            label=f'[mvThemeComponent] color ({name})',
        )
        style_c = mvThemeComponent.aliased(
            tag=f'[mvThemeComponent] style ({name})',
            label=f'[mvThemeComponent] style ({name})',
        )
    try:
        yield theme
    finally:
        id_element = api.ThemeElement.identify
        for elem_tp, c in (('mvThemeColor', color_c), ('mvThemeStyle', style_c)):
            label_pfx = elem_tp.removeprefix('mvTheme').lower()
            for e in c.children(1):
                alias = f'[{elem_tp}] {id_element(e)} ({name}.{label_pfx})'
                api.Item.set_alias(
                    e,
                    alias,
                )
                api.Item.configure(e, label=alias)




def load_presets() -> tuple[mvTheme, ...]:
    """Return all Dear PyPixl presets, loading them if they do
    not exist.
    """
    return tuple(fn() for fn in _preset_fns)




# [ THEMES ]

# "Complete" themes use most but not all elements. The
# functions that make the unused elements are simply left
# uncalled; this makes it easier to tell which elements are
# missing at a glance.


@_dearpypixl_preset('Default')
def default():
    """Return the preset that emulates Dear PyGui's internal
    default theme.
    """
    with _theme_preset('Default') as theme:
        with mvThemeComponent.new(theme[1][0]):
            color.border(78, 78, 78, 255)
            color.border_shadow(78, 78, 78, 255)
            color.button(51, 51, 55, 255)
            color.button_active(0, 119, 200, 153)
            color.button_hovered(29, 151, 236, 103)
            color.check_mark(0, 119, 200, 153)
            color.child_bg(37, 37, 38, 255)
            color.docking_empty_bg(51, 51, 51, 255)
            color.docking_preview(0, 119, 200, 153)
            color.drag_drop_target(255, 255, 0, 179)
            color.frame_bg(51, 51, 55, 255)
            color.frame_bg_active(0, 119, 200, 153)
            color.frame_bg_hovered(29, 151, 236, 103)
            color.header(51, 51, 55, 255)
            color.header_active(0, 119, 200, 153)
            color.header_hovered(29, 151, 236, 103)
            color.menu_bar_bg(51, 51, 55, 255)
            color.modal_window_dim_bg(37, 37, 38, 150)
            color.nav_highlight(37, 37, 38, 255)
            color.nav_windowing_dim_bg(204, 204, 204, 51)
            color.nav_windowing_highlight(255, 255, 255, 179)
            color.popup_bg(37, 37, 38, 255)
            color.resize_grip(37, 37, 38, 255)
            color.resize_grip_hovered(51, 51, 55, 255)
            color.resize_grip_active(51, 51, 55, 255)
            color.scrollbar_bg(51, 51, 55, 255)
            color.scrollbar_grab(82, 82, 85, 255)
            color.scrollbar_grab_active(90, 90, 95, 255)
            color.scrollbar_grab_hovered(90, 90, 95, 255)
            color.separator(78, 78, 78, 255)
            color.separator_active(78, 78, 78, 255)
            color.separator_hovered(78, 78, 78, 255)
            color.slider_grab(29, 151, 236, 103)
            color.slider_grab_active(0, 119, 200, 153)
            color.tab(51, 51, 55, 255)
            color.tab_active(0, 119, 200, 153)
            color.tab_hovered(29, 151, 236, 103)
            color.tab_unfocused(51, 51, 55, 255)
            color.tab_unfocused_active(0, 119, 200, 153)
            color.table_border_light(59, 59, 64, 255)
            color.table_border_strong(79, 79, 89, 255)
            color.table_header_bg(48, 48, 51, 255)
            color.table_row_bg(0, 0, 0, 0)
            color.table_row_bg_alt(255, 255, 255, 15)
            color.text(255, 255, 255, 255)
            color.text_disabled(151, 151, 151, 255)
            color.text_selected_bg(0, 119, 200, 153)
            color.title_bg(37, 37, 38, 255)
            color.title_bg_active(15, 86, 135, 255)
            color.title_bg_collapsed(37, 37, 38, 255)
            color.window_bg(37, 37, 38, 255)
            color.plot_bg
            color.plot_border
            color.plot_crosshairs
            color.plot_error_bar
            color.plot_fill
            color.plot_frame_bg
            color.plot_histogram(0, 119, 200, 153)
            color.plot_histogram_hovered(29, 151, 236, 103)
            color.plot_inlay_text
            color.plot_legend_bg
            color.plot_legend_border
            color.plot_legend_text
            color.plot_line
            color.plot_lines(0, 119, 200, 153)
            color.plot_lines_hovered(29, 151, 236, 103)
            color.plot_marker_fill
            color.plot_marker_outline
            color.plot_query
            color.plot_selection
            color.plot_title_text
            color.plot_x_axis
            color.plot_x_axis_grid
            color.plot_y_axis
            color.plot_y_axis2
            color.plot_y_axis3
            color.plot_y_axis_grid
            color.plot_y_axis_grid2
            color.plot_y_axis_grid3
            color.node_bg(62, 62, 62, 255)
            color.node_bg_hovered(75, 75, 75, 255)
            color.node_bg_selected(75, 75, 75, 255)
            color.node_box_selector(61, 133, 224, 30)
            color.node_box_selector_outline(61, 133, 224, 150)
            color.node_grid_bg(35, 35, 35, 255)
            color.node_grid_line(0, 0, 0, 255)
            color.node_grid_line_primary(240, 240, 240, 60)
            color.node_link(255, 255, 255, 200)
            color.node_link_hovered(66, 150, 250, 255)
            color.node_link_selected(66, 150, 250, 255)
            color.node_mini_map_bg(25, 25, 25, 150)
            color.node_mini_map_bg_hovered(150, 150, 150, 200)
            color.node_mini_map_canvas(200, 200, 200, 25)
            color.node_mini_map_canvas_outline(200, 200, 200, 200)
            color.node_mini_map_link(61, 133, 224, 200)
            color.node_mini_map_link_selected(66, 150, 250, 255)
            color.node_mini_map_node_bg(200, 200, 200, 100)
            color.node_mini_map_node_bg_hovered(200, 200, 200, 255)
            color.node_mini_map_node_bg_selected(200, 200, 200, 255)
            color.node_mini_map_node_outline(200, 200, 200, 100)
            color.node_mini_map_outline(150, 150, 150, 100)
            color.node_mini_map_outline_hovered(150, 150, 150, 200)
            color.node_outline(100, 100, 100, 255)
            color.node_pin(199, 199, 41, 255)
            color.node_pin_hovered(255, 255, 50, 255)
            color.node_title_bar(37, 37, 38, 255)
            color.node_title_bar_hovered(15, 86, 135, 255)
            color.node_title_bar_selected(0, 119, 200, 153)
        with mvThemeComponent.new(theme[1][1]):
            style.alpha
            style.button_text_align(0.5, 0.5)
            style.cell_padding(4.0, 2.0)
            style.child_border_size(1)
            style.child_rounding(0)
            style.frame_border_size(0)
            style.frame_padding(4.0, 3.0)
            style.frame_rounding(0)
            style.grab_min_size(10.0)
            style.grab_rounding(0)
            style.indent_spacing(21.0)
            style.item_inner_spacing(4, 4)
            style.item_spacing(8, 4)
            style.popup_border_size(1)
            style.popup_rounding(0)
            style.scrollbar_rounding(9)
            style.scrollbar_size(14)
            style.selectable_text_align(0.0, 0.5)
            style.tab_rounding(4)
            style.window_border_size(1)
            style.window_min_size(100, 100)
            style.window_padding(8, 8)
            style.window_rounding(0)
            style.window_title_align(0.0, 0.5)
            style.plot_annotation_padding(2.0, 2.0)
            style.plot_border_size(1)
            style.plot_default_size(400, 300)
            style.plot_digital_bit_gap(8.0)
            style.plot_digital_bit_height(4.0)
            style.plot_error_bar_size(5.0)
            style.plot_error_bar_weight(1.0)
            style.plot_fill_alpha(1.0)
            style.plot_fit_padding(0.0, 0.0)
            style.plot_label_padding(5, 5)
            style.plot_legend_inner_padding(5, 5)
            style.plot_legend_padding(10, 10)
            style.plot_legend_spacing(5, 0)
            style.plot_line_weight(1.0)
            style.plot_major_grid_size(1.0, 1.0)
            style.plot_major_tick_len(10, 10)
            style.plot_major_tick_size(1.0, 1.0)
            style.plot_marker(4.0)
            style.plot_marker_size(4.0)
            style.plot_marker_weight(1.0)
            style.plot_min_size(200, 150)
            style.plot_minor_alpha(0.25)
            style.plot_minor_grid_size(1.0, 1.0)
            style.plot_minor_tick_len(5, 5)
            style.plot_minor_tick_size(1.0, 1.0)
            style.plot_mouse_pos_padding(10, 10)
            style.plot_padding(10, 10)
            style.node_border_thickness(1)
            style.node_corner_rounding(4)
            style.node_grid_spacing(24)
            style.node_link_hover_distance(10)
            style.node_link_line_segments_per_length(0)
            style.node_link_thickness(3)
            #style.node_mini_map_offset(4, 0)  DPG BUG?
            style.node_mini_map_padding(8, 8)
            style.node_padding(8, 8)
            style.node_pin_circle_radius(4)
            style.node_pin_hover_radius(10)
            style.node_pin_line_thickness(1)
            style.node_pin_offset(0)
            style.node_pin_quad_side_length(7)
            style.node_pin_triangle_side_length(10)
    return theme


@_dearpypixl_preset('ImGui Dark')
def imgui_dark():
    """Return the preset emulating `dearpygui_ext`s "imgui dark"
    theme.
    """
    with _theme_preset('ImGui Dark') as theme:
        with mvThemeComponent.new(theme[1][0]):
            color.border(110, 110, 128, 128)
            color.border_shadow(0, 0, 0, 0)
            color.button(66, 150, 250, 102)
            color.button_active(15, 135, 250, 255)
            color.button_hovered(66, 150, 250, 255)
            color.check_mark(66, 150, 250, 255)
            color.child_bg(0, 0, 0, 0)
            color.docking_empty_bg(51, 51, 51, 255)
            color.docking_preview(66, 150, 250, 178)
            color.drag_drop_target(255, 255, 0, 230)
            color.frame_bg(41, 74, 122, 138)
            color.frame_bg_active(66, 150, 250, 171)
            color.frame_bg_hovered(66, 150, 250, 102)
            color.header(66, 150, 250, 79)
            color.header_active(66, 150, 250, 255)
            color.header_hovered(66, 150, 250, 204)
            color.menu_bar_bg(36, 36, 36, 255)
            color.modal_window_dim_bg(204, 204, 204, 89)
            color.nav_highlight(66, 150, 250, 255)
            color.nav_windowing_dim_bg(204, 204, 204, 51)
            color.nav_windowing_highlight(255, 255, 255, 178)
            color.popup_bg(20, 20, 20, 240)
            color.resize_grip(66, 150, 250, 51)
            color.resize_grip_active(66, 150, 250, 242)
            color.resize_grip_hovered(66, 150, 250, 171)
            color.scrollbar_bg(5, 5, 5, 135)
            color.scrollbar_grab(79, 79, 79, 255)
            color.scrollbar_grab_active(130, 130, 130, 255)
            color.scrollbar_grab_hovered(105, 105, 105, 255)
            color.separator(110, 110, 128, 128)
            color.separator_active(26, 102, 191, 255)
            color.separator_hovered(26, 102, 191, 199)
            color.slider_grab(61, 133, 224, 255)
            color.slider_grab_active(66, 150, 250, 255)
            color.tab(46, 89, 148, 219)
            color.tab_active(51, 105, 173, 255)
            color.tab_hovered(66, 150, 250, 204)
            color.tab_unfocused(18, 26, 38, 247)
            color.tab_unfocused_active(36, 66, 107, 255)
            color.table_border_light(59, 59, 64, 255)
            color.table_border_strong(79, 79, 89, 255)
            color.table_header_bg(48, 48, 51, 255)
            color.table_row_bg(0, 0, 0, 0)
            color.table_row_bg_alt(255, 255, 255, 15)
            color.text(255, 255, 255, 255)
            color.text_disabled(128, 128, 128, 255)
            color.text_selected_bg(66, 150, 250, 89)
            color.title_bg(10, 10, 10, 255)
            color.title_bg_active(41, 74, 122, 255)
            color.title_bg_collapsed(0, 0, 0, 130)
            color.window_bg(15, 15, 15, 240)
            color.plot_bg(0, 0, 0, 128)
            color.plot_border(110, 110, 128, 128)
            color.plot_crosshairs(255, 255, 255, 128)
            color.plot_error_bar
            color.plot_fill
            color.plot_frame_bg(255, 255, 255, 18)
            color.plot_histogram(230, 178, 0, 255)
            color.plot_histogram_hovered(255, 153, 0, 255)
            color.plot_inlay_text(255, 255, 255, 255)
            color.plot_legend_bg(20, 20, 20, 240)
            color.plot_legend_border(110, 110, 128, 128)
            color.plot_legend_text(255, 255, 255, 255)
            color.plot_line
            color.plot_lines(156, 156, 156, 255)
            color.plot_lines_hovered(255, 110, 89, 255)
            color.plot_marker_fill
            color.plot_marker_outline
            color.plot_query(0, 255, 112, 255)
            color.plot_selection(255, 153, 0, 255)
            color.plot_title_text(255, 255, 255, 255)
            color.plot_x_axis(255, 255, 255, 255)
            color.plot_x_axis_grid(255, 255, 255, 64)
            color.plot_y_axis(255, 255, 255, 255)
            color.plot_y_axis2(255, 255, 255, 255)
            color.plot_y_axis3(255, 255, 255, 255)
            color.plot_y_axis_grid(255, 255, 255, 64)
            color.plot_y_axis_grid2(255, 255, 255, 64)
            color.plot_y_axis_grid3(255, 255, 255, 64)
            color.node_bg(50, 50, 50, 255)
            color.node_bg_hovered(75, 75, 75, 255)
            color.node_bg_selected(75, 75, 75, 255)
            color.node_box_selector(61, 133, 224, 30)
            color.node_box_selector_outline(61, 133, 224, 150)
            color.node_grid_bg(40, 40, 50, 200)
            color.node_grid_line(200, 200, 200, 40)
            color.node_grid_line_primary
            color.node_link(61, 133, 224, 200)
            color.node_link_hovered(66, 150, 250, 255)
            color.node_link_selected(66, 150, 250, 255)
            color.node_mini_map_bg
            color.node_mini_map_bg_hovered
            color.node_mini_map_canvas
            color.node_mini_map_canvas_outline
            color.node_mini_map_link
            color.node_mini_map_link_selected
            color.node_mini_map_node_bg
            color.node_mini_map_node_bg_hovered
            color.node_mini_map_node_bg_selected
            color.node_mini_map_node_outline
            color.node_mini_map_outline
            color.node_mini_map_outline_hovered
            color.node_outline(100, 100, 100, 255)
            color.node_pin(53, 150, 250, 180)
            color.node_pin_hovered(53, 150, 250, 255)
            color.node_title_bar(41, 74, 122, 255)
            color.node_title_bar_hovered(66, 150, 250, 255)
            color.node_title_bar_selected(66, 150, 250, 255)
        with mvThemeComponent.new(theme[1][1]):
            style.alpha
            style.button_text_align(0.5, 0.5)
            style.cell_padding(4.0, 2.0)
            style.child_border_size(1)
            style.child_rounding(0)
            style.frame_border_size(0)
            style.frame_padding(4.0, 3.0)
            style.frame_rounding(0)
            style.grab_min_size(10.0)
            style.grab_rounding(0)
            style.indent_spacing(21.0)
            style.item_inner_spacing(4, 4)
            style.item_spacing(8, 4)
            style.popup_border_size(1)
            style.popup_rounding(0)
            style.scrollbar_rounding(9)
            style.scrollbar_size(14)
            style.selectable_text_align(0.0, 0.5)
            style.tab_rounding(4)
            style.window_border_size(1)
            style.window_min_size(100, 100)
            style.window_padding(8, 8)
            style.window_rounding(0)
            style.window_title_align(0.0, 0.5)
            style.plot_annotation_padding(2.0, 2.0)
            style.plot_border_size(1)
            style.plot_default_size(400, 300)
            style.plot_digital_bit_gap(8.0)
            style.plot_digital_bit_height(4.0)
            style.plot_error_bar_size(5.0)
            style.plot_error_bar_weight(1.0)
            style.plot_fill_alpha(1.0)
            style.plot_fit_padding(0.0, 0.0)
            style.plot_label_padding(5, 5)
            style.plot_legend_inner_padding(5, 5)
            style.plot_legend_padding(10, 10)
            style.plot_legend_spacing(5, 0)
            style.plot_line_weight(1.0)
            style.plot_major_grid_size(1.0, 1.0)
            style.plot_major_tick_len(10, 10)
            style.plot_major_tick_size(1.0, 1.0)
            style.plot_marker(4.0)
            style.plot_marker_size(4.0)
            style.plot_marker_weight(1.0)
            style.plot_min_size(200, 150)
            style.plot_minor_alpha(0.25)
            style.plot_minor_grid_size(1.0, 1.0)
            style.plot_minor_tick_len(5, 5)
            style.plot_minor_tick_size(1.0, 1.0)
            style.plot_mouse_pos_padding(10, 10)
            style.plot_padding(10, 10)
            style.node_border_thickness(1)
            style.node_corner_rounding(4)
            style.node_grid_spacing(24)
            style.node_link_hover_distance(10)
            style.node_link_line_segments_per_length(0)
            style.node_link_thickness(3)
            #style.node_mini_map_offset(4, 0)  DPG BUG?
            style.node_mini_map_padding(8, 8)
            style.node_padding(8, 8)
            style.node_pin_circle_radius(4)
            style.node_pin_hover_radius(10)
            style.node_pin_line_thickness(1)
            style.node_pin_offset(0)
            style.node_pin_quad_side_length(7)
            style.node_pin_triangle_side_length(10)
    return theme


# TODO
# @_dearpypixl_preset('ImGui Light')
def _imgui_light():
    """Return the preset emulating `dearpygui_ext`s "imgui light"
    theme.
    """
    with _theme_preset('ImGui Light') as theme:
        with mvThemeComponent.new(theme[1][0]):
            color.border
            color.border_shadow
            color.button
            color.button_active
            color.button_hovered
            color.check_mark
            color.child_bg
            color.docking_empty_bg
            color.docking_preview
            color.drag_drop_target
            color.frame_bg
            color.frame_bg_active
            color.frame_bg_hovered
            color.header
            color.header_active
            color.header_hovered
            color.menu_bar_bg
            color.modal_window_dim_bg
            color.nav_highlight
            color.nav_windowing_dim_bg
            color.nav_windowing_highlight
            color.popup_bg
            color.resize_grip
            color.resize_grip_active
            color.resize_grip_hovered
            color.scrollbar_bg
            color.scrollbar_grab
            color.scrollbar_grab_active
            color.scrollbar_grab_hovered
            color.separator
            color.separator_active
            color.separator_hovered
            color.slider_grab
            color.slider_grab_active
            color.tab
            color.tab_active
            color.tab_hovered
            color.tab_unfocused
            color.tab_unfocused_active
            color.table_border_light
            color.table_border_strong
            color.table_header_bg
            color.table_row_bg
            color.table_row_bg_alt
            color.text
            color.text_disabled
            color.text_selected_bg
            color.title_bg
            color.title_bg_active
            color.title_bg_collapsed
            color.window_bg
            color.plot_bg
            color.plot_border
            color.plot_crosshairs
            color.plot_error_bar
            color.plot_fill
            color.plot_frame_bg
            color.plot_histogram
            color.plot_histogram_hovered
            color.plot_inlay_text
            color.plot_legend_bg
            color.plot_legend_border
            color.plot_legend_text
            color.plot_line
            color.plot_lines
            color.plot_lines_hovered
            color.plot_marker_fill
            color.plot_marker_outline
            color.plot_query
            color.plot_selection
            color.plot_title_text
            color.plot_x_axis
            color.plot_x_axis_grid
            color.plot_y_axis
            color.plot_y_axis2
            color.plot_y_axis3
            color.plot_y_axis_grid
            color.plot_y_axis_grid2
            color.plot_y_axis_grid3
            color.node_bg
            color.node_bg_hovered
            color.node_bg_selected
            color.node_box_selector
            color.node_box_selector_outline
            color.node_grid_bg
            color.node_grid_line
            color.node_grid_line_primary
            color.node_link
            color.node_link_hovered
            color.node_link_selected
            color.node_mini_map_bg
            color.node_mini_map_bg_hovered
            color.node_mini_map_canvas
            color.node_mini_map_canvas_outline
            color.node_mini_map_link
            color.node_mini_map_link_selected
            color.node_mini_map_node_bg
            color.node_mini_map_node_bg_hovered
            color.node_mini_map_node_bg_selected
            color.node_mini_map_node_outline
            color.node_mini_map_outline
            color.node_mini_map_outline_hovered
            color.node_outline
            color.node_pin
            color.node_pin_hovered
            color.node_title_bar
            color.node_title_bar_hovered
            color.node_title_bar_selected
        with mvThemeComponent.new(theme[1][1]):
            style.alpha
            style.button_text_align(0.5, 0.5)
            style.cell_padding(4.0, 2.0)
            style.child_border_size(1)
            style.child_rounding(0)
            style.frame_border_size(0)
            style.frame_padding(4.0, 3.0)
            style.frame_rounding(0)
            style.grab_min_size(10.0)
            style.grab_rounding(0)
            style.indent_spacing(21.0)
            style.item_inner_spacing(4, 4)
            style.item_spacing(8, 4)
            style.popup_border_size(1)
            style.popup_rounding(0)
            style.scrollbar_rounding(9)
            style.scrollbar_size(14)
            style.selectable_text_align(0.0, 0.5)
            style.tab_rounding(4)
            style.window_border_size(1)
            style.window_min_size(100, 100)
            style.window_padding(8, 8)
            style.window_rounding(0)
            style.window_title_align(0.0, 0.5)
            style.plot_annotation_padding(2.0, 2.0)
            style.plot_border_size(1)
            style.plot_default_size(400, 300)
            style.plot_digital_bit_gap(8.0)
            style.plot_digital_bit_height(4.0)
            style.plot_error_bar_size(5.0)
            style.plot_error_bar_weight(1.0)
            style.plot_fill_alpha(1.0)
            style.plot_fit_padding(0.0, 0.0)
            style.plot_label_padding(5, 5)
            style.plot_legend_inner_padding(5, 5)
            style.plot_legend_padding(10, 10)
            style.plot_legend_spacing(5, 0)
            style.plot_line_weight(1.0)
            style.plot_major_grid_size(1.0, 1.0)
            style.plot_major_tick_len(10, 10)
            style.plot_major_tick_size(1.0, 1.0)
            style.plot_marker(4.0)
            style.plot_marker_size(4.0)
            style.plot_marker_weight(1.0)
            style.plot_min_size(200, 150)
            style.plot_minor_alpha(0.25)
            style.plot_minor_grid_size(1.0, 1.0)
            style.plot_minor_tick_len(5, 5)
            style.plot_minor_tick_size(1.0, 1.0)
            style.plot_mouse_pos_padding(10, 10)
            style.plot_padding(10, 10)
            style.node_border_thickness(1)
            style.node_corner_rounding(4)
            style.node_grid_spacing(24)
            style.node_link_hover_distance(10)
            style.node_link_line_segments_per_length(0)
            style.node_link_thickness(3)
            #style.node_mini_map_offset(4, 0)  DPG BUG?
            style.node_mini_map_padding(8, 8)
            style.node_padding(8, 8)
            style.node_pin_circle_radius(4)
            style.node_pin_hover_radius(10)
            style.node_pin_line_thickness(1)
            style.node_pin_offset(0)
            style.node_pin_quad_side_length(7)
            style.node_pin_triangle_side_length(10)
    return theme


@_dearpypixl_preset("Padless")
def padless():
    with _theme_preset("Padless") as theme:
        with mvThemeComponent.new(theme[1][1]):
            style.cell_padding(0, 0)
            style.item_spacing(0, 0)
            style.window_padding(0, 0)
    return theme



"""Callback and event-related utilities for Dear PyPixl
and Dear PyGui."""
import time
import enum
import inspect
import functools
import threading
import collections
from inspect import Parameter as _Parameter
from dearpygui import dearpygui
from .constants import KeyInput, MouseInput
from . import (
    api,
    _interface,
    _tools,
)
# BUG: DPG leaks callback arguments. `user_data` is leaked
# in some cases, but `app_data` leaks are plentiful and can
# be a BIG problem. Non-container values are sometimes fine;
# even if these do leak, they're not the primary concern.
# Containters though...we decref those. We decref those to
# *heck*.
from ._tools import Py_DECREF
from ._typing import (
    Any,
    Item,
    ItemCommand,
    ItemCallback,
    Array,
    ItemConfig,
    Property,
    Iterable,
    Collection,
    Callable,
    MethodType,
    SupportsIndex,
    ItemInterface,
    TypeVar,
    ParamSpec,
    Self,
    overload,
    override,
    cast,
)
from . import items


_T   = TypeVar("_T")
_N   = TypeVar("_N", float, int)
_P   = ParamSpec("_P")




class _empty: ... # this declaration keeps the typechecker quiet
_empty = _tools.create_marker_type("Empty")  # type: ignore

Empty = type[_empty]







# [ HANDLER REGISTRY EXTENSIONS ]

# handler signatures -- protocols are just more work here

def _handler(
    self,
    callback: ItemCallback | None = None,
    *,
    label             : str | None = None,
    user_data         : Any        = None,
    use_internal_label: bool       = True,
    tag               : Item       = 0,
    parent            : Item       = 0,
    show              : bool       = True,
    **kwargs
) -> ItemCallback | None: ...
def _mouse_handler(
    self,
    callback: ItemCallback | None = None,
    *,
    button            : int        = -1,
    label             : str | None = None,
    user_data         : Any        = None,
    use_internal_label: bool       = True,
    tag               : Item       = 0,
    parent            : Item       = 0,
    show              : bool       = True,
    **kwargs
) -> ItemCallback | None: ...
def _key_hander(
    self,
    callback: ItemCallback | None = None,
    *,
    key               : int        = -1,
    label             : str | None = None,
    user_data         : Any        = None,
    use_internal_label: bool       = True,
    tag               : Item       = 0,
    parent            : Item       = 0,
    show              : bool       = True,
    **kwargs
) -> ItemCallback | None: ...


def handler_hook(handler_fn: Any, protocol: Callable[_P, _T] = _handler) -> Callable[[Callable], Callable[_P, _T]]:
    def wrap_method(mthd: Any) -> Callable[_P, _T]:
        @functools.wraps(mthd)
        def mthd_add_handler(self, callback = None, *args, parent: Item = 0, **kwargs):
            def add_handler(callback):
                handler_fn(
                    callback=callback,
                    parent=parent or self,
                    **kwargs
                )
                return callback

            if callback is None:
                return add_handler
            return add_handler(callback)

        return mthd_add_handler  # type: ignore

    return wrap_method




class HandlerRegistry(items.mvHandlerRegistry):
    """`mvHandlerRegistry` extension exposing global input
    handler methods, which can optionally be used as decorators.

    The signature of each method slightly differs from the
    DearPyGui command hook used. For all methods, the *callback*
    parameter is the only positional argument. All other arguments
    are optional and keyword-only. Each method returns the
    *callback* argument.

    Handler methods can be used as decorators both with and without
    parenthesis (you do not need to call them). Using parenthesis
    will allow for passing arguments; in this case, you should not
    include a *callback* argument.

    Command: `dearpygui.add_handler_registry`
    """

    @handler_hook(dearpygui.add_mouse_click_handler, _mouse_handler)
    def on_mouse_click(self, *args, **kwargs):
        """Schedule a callback to run when both a mouse button
        down and up event occur on the same object.

        Command: `dearpygui.add_mouse_click_handler`
        """

    @handler_hook(dearpygui.add_mouse_down_handler, _mouse_handler)
    def on_mouse_click_down(self, *args, **kwargs):
        """Schedule a callback to run when a mouse button is
        clicked/pressed.

        Command: `dearpygui.add_mouse_down_handler`
        """

    @handler_hook(dearpygui.add_mouse_release_handler, _mouse_handler)
    def on_mouse_click_up(self, *args, **kwargs):
        """Schedule a callback to run when a mouse button is
        released.

        Command: `dearpygui.add_mouse_release_handler`
        """

    @handler_hook(dearpygui.add_mouse_double_click_handler, _mouse_handler)
    def on_mouse_double_click(self, *args, **kwargs):
        """Schedule a callback to run when a mouse click event
        occurs twice consecutively on the same object.

        Command: `dearpygui.add_mouse_double_click_handler`
        """

    @handler_hook(dearpygui.add_mouse_wheel_handler, _mouse_handler)
    def on_mouse_wheel(self, *args, **kwargs):
        """Schedule a callback to run on mouse wheel input
        (excluding middle click).

        Command: `dearpygui.add_mouse_wheel_handler`
        """

    @handler_hook(dearpygui.add_mouse_move_handler, _mouse_handler)
    def on_mouse_move(self, *args, **kwargs):
        """Schedule a callback to run when a mouse is moved.

        Command: `dearpygui.add_mouse_move_handler`
        """

    @handler_hook(dearpygui.add_mouse_drag_handler, _mouse_handler)
    def on_mouse_drag(self, *args, **kwargs):
        """Schedule a callback to run when a mouse move event
        occurs during a mouse click down event.

        Command: `dearpygui.add_mouse_drag_handler`
        """

    @handler_hook(dearpygui.add_key_down_handler, _key_hander)
    def on_key_down(self, *args, **kwargs):
        """Schedule a callback to run on key down input. On
        many OS, this will fire continuously while the key is
        down.

        Command: `dearpygui.add_key_down_handler`

        NOTE: 'Key down' and 'key press' events are similar
        but not identical -- A 'key down' event occurs before
        a 'key press' event.
        """

    @handler_hook(dearpygui.add_key_press_handler, _key_hander)
    def on_key_press(self, *args, **kwargs):
        """Schedule a callback to run when a key is pressed.
        On many OS, this will fire continuously while the key
        is pressed.

        Command: `dearpygui.add_key_press_handler`

        NOTE: 'Key down' and 'key press' events are similar
        but not identical -- A 'key press' event occurs after
        a 'key down' event.
        """

    @handler_hook(dearpygui.add_key_release_handler, _key_hander)
    def on_key_up(self, *args, **kwargs):
        """Schedule a callback to run when a key is released.

        Command: `dearpygui.add_key_release_handler`
        """


class ItemHandlerRegistry(items.mvItemHandlerRegistry):
    """`mvItemHandlerRegistry` extension exposing item handler
    as methods, which can optionally be used as decorators.

    The signature of each method slightly differs from the
    DearPyGui command hook used. For all methods, the *callback*
    parameter is the only positional argument. All other arguments
    are optional and keyword-only. Each method returns the
    *callback* argument.

    Handler methods can be used as decorators both with and without
    parenthesis (you do not need to call them). Using parenthesis
    will allow for passing arguments; in this case, you should not
    include a *callback* argument.

    Command: `dearpygui.add_item_handler_registry`
    """

    @handler_hook(dearpygui.add_item_resize_handler)
    def on_resize(self, *args, **kwargs):
        """Schedule a callback to run when an item bound to
        this registry is resized.

        Command: `dearpygui.add_item_resize_handler`
        """

    @handler_hook(dearpygui.add_item_clicked_handler, _mouse_handler)
    def on_click(self, *args, **kwargs):
        """Schedule a callback to run when both a mouse button
        down and button up events occur consecutively on a single
        item bound to this registry.

        Command: `dearpygui.add_item_clicked_handler`
        """

    @handler_hook(dearpygui.add_item_toggled_open_handler)
    def on_toggle_open(self, *args, **kwargs):
        """Schedule a callback to run when an item bound to this
        registry is toggled open.

        Command: `dearpygui.add_item_toggled_open_handler`
        """

    @handler_hook(dearpygui.add_item_edited_handler)
    def on_edit(self, *args, **kwargs):
        """Schedule a callback to run when an item bound to this
        registry is edited.

        Command: `dearpygui.add_item_edited_handler`
        """

    @handler_hook(dearpygui.add_item_deactivated_after_edit_handler)
    def on_deactivation_after_edit(self, *args, **kwargs):
        """Schedule a callback to run when an item bound to this
        registry is deactivated and
        recently edited.

        Command: `dearpygui.add_item_deactivated_after_edit_handler`
        """

    @handler_hook(dearpygui.add_item_activated_handler)
    def on_activation(self, *args, **kwargs):
        """Schedules a callback to run when an item bound to this
        registry is interacted with.

        Command: `dearpygui.add_item_activated_handler`
        """

    @handler_hook(dearpygui.add_item_deactivated_handler)
    def on_deactivation(self, *args, **kwargs):
        """Schedule a callback to run when an item bound to this
        registry is deactivated.

        Command: `dearpygui.add_item_deactivated_handler`
        """

    @handler_hook(dearpygui.add_item_active_handler)
    def while_enabled(self, *args, **kwargs):
        """Schedules a callback to run when an item bound to this
        registry is not disabled.

        Command: `dearpygui.add_item_active_handler`
        """

    @handler_hook(dearpygui.add_item_visible_handler)
    def while_visible(self, *args, **kwargs):
        """Schedules a callback to run when an item bound to
        this registry is visible.

        Command: `dearpygui.add_item_visible_handler`
        """

    @handler_hook(dearpygui.add_item_focus_handler)
    def while_focused(self, *args, **kwargs):
        """Schedules a callback to run when an item bound to
        this registry is focused.

        Command: `dearpygui.add_item_focus_handler`
        """

    @handler_hook(dearpygui.add_item_hover_handler)
    def while_hovered(self, *args, **kwargs):
        """Schedules a callback to run when an item bound to
        this registry is hovered.

        Command: `dearpygui.add_item_hover_handler`
        """




# enum conversions are a little slow, so the associations
# are cached for quicker access
_KEYCODE_MAP = {m.value: m for m in KeyInput}
_PTRCODE_MAP = {m.value: m for m in MouseInput}

class _InputPoller(items.mvHandlerRegistry):
    def __init__(self, *args, **kwargs) -> None:
        if not 'label' in kwargs:
            kwargs['label'] = f'[{type(self).__name__}]'
        super().__init__(*args, **kwargs)
        # Most of the traffic will be from the DPG callback
        # thread, so the necessity of this lock is questionable.
        self.lock = threading.RLock()
        self._inputs = []

    def __iter__(self):
        with self.lock:
            return iter(self._inputs)

    def __len__(self):
        with self.lock:
            return len(self._inputs)

    def __str__(self):
        return str(self._inputs)

    @overload
    def __getitem__(self, index: SupportsIndex): ...
    @overload
    def __getitem__(self, index: slice): ...
    def __getitem__(self, index: Any):
        return self._inputs[index]

    def __contains__(self, other: Any):
        return other in self._inputs


class InputEvent(enum.StrEnum):
    NONE           = "NONE"
    KEY_DOWN       = "KEY_DOWN"
    KEY_UP         = "KEY_UP"
    KEY_PRESS      = "KEY_PRESS"
    MOUSE_DOWN     = "MOUSE_DOWN"
    MOUSE_UP       = "MOUSE_UP"
    MOUSE_MOVE     = "MOUSE_MOVE"
    MOUSE_DRAG     = "MOUSE_DRAG"
    MOUSE_V_SCROLL = "MOUSE_V_SCROLL"


class InputRecorder(_InputPoller):
    """A global handler registry that records input events.
    Useful for logging and debugging.

    The registry behaves like a primitive sequence; supporting
    iteration, indexing, and slicing. Iterating the registry
    yields 3-item tuples containing polled input data. The first
    item is the approx. time of the event as returned by
    `time.time()`. The second indicates the type of input that
    was captured, and is always a member of the `InputEvent`
    string enumeration. The last item is the value Dear PyGui
    sent as `app_data` to the related handler's callback when
    the input event triggered. The value type of the third item
    varies depending on the type of input captured;
        * `InputEvent.KEY_DOWN`: A 2-item list containing the key
        event code and the duration it has been held down in
        fractional seconds.

        * `InputEvent.KEY_UP`: Event code of the key released.

        * `InputEvent.MOUSE_DOWN`: A 2-item tuple containing the
        button event code and the duration it has been held down
        in fractional seconds.

        * `InputEvent.MOUSE_UP`: Event code of the button released.

        * `InputEvent.MOUSE_MOVE`: A 2-item tuple containing the
        x and y position of the cursor relative to the top-left
        corner of the Dear PyGui viewport.

        * `InputEvent.MOUSE_DRAG`: A 3-item tuple containing the
        event code of the button held down and the x/y position
        delta of the cursor. Coordinates are relative to the
        location of the cursor when the button was first pressed.

        * `InputEvent.MOUSE_V_SCROLL`: A single integer indicating
        the direction and intensity of the scroll input. The value
        is positive when a "scroll up" input is captured, and vice-
        versa. The speed or intensity of the input is indicated by
        the integer's absolute value; the higher the value, the
        greater the intensity.

        * `InputEvent.NONE`: Indicates the instantiation of the
        object, and is always the first event "recorded". The
        value is always 0.

    "Push"/"click" events are captured as individual "down"
    and "up" events.
    """

    KEY_DOWN = InputEvent.KEY_DOWN

    def _cb_key_down(self, handler: Item, key_info: tuple[int, float]):
        with self.lock:
            self._inputs.append(
                (time.time(), self.KEY_DOWN, (_KEYCODE_MAP[key_info[0]], key_info[1]))
            )
            Py_DECREF(key_info)


    MOUSE_DOWN = InputEvent.MOUSE_DOWN

    def _cb_mouse_down(self, handler: Item, key_info: tuple[int, float]):
        with self.lock:
            self._inputs.append(
                (time.time(), self.MOUSE_DOWN, (_PTRCODE_MAP[key_info[0]], key_info[1]))
            )
            Py_DECREF(key_info)


    KEY_UP = InputEvent.KEY_UP

    def _cb_key_up(self, handler: Item, key: int):
        with self.lock:
            self._inputs.append((time.time(), self.KEY_UP, _KEYCODE_MAP[key]))


    MOUSE_UP = InputEvent.MOUSE_UP

    def _cb_mouse_up(self, handler: Item, key: int):
        with self.lock:
            self._inputs.append((time.time(), self.MOUSE_UP, _PTRCODE_MAP[key]))


    MOUSE_MOVE = InputEvent.MOUSE_MOVE

    def _cb_mouse_move(self, handler:  Item, cursor_pos: tuple[float, float]):
        with self.lock:
            self._inputs.append((time.time(), self.MOUSE_MOVE, tuple(cursor_pos)))
            Py_DECREF(cursor_pos)


    MOUSE_DRAG = InputEvent.MOUSE_DRAG

    def _cb_mouse_drag(self, handler: Item, drag_data: tuple[int, float, float]):
        key, x_pos_dt, y_pos_dt = drag_data
        with self.lock:
            self._inputs.append((time.time(), self.MOUSE_DRAG, (_PTRCODE_MAP[key], x_pos_dt, y_pos_dt)))
            Py_DECREF(drag_data)


    MOUSE_V_SCROLL = InputEvent.MOUSE_V_SCROLL

    def _cb_mouse_v_scroll(self, handler: Item, value: int):
        with self.lock:
            self._inputs.append((time.time(), self.MOUSE_V_SCROLL, value))


    __INPUT_HOOK_MAP = {
        KEY_DOWN      : items.mvKeyDownHandler,
        KEY_UP        : items.mvKeyReleaseHandler,
        MOUSE_DOWN    : items.mvMouseDownHandler,
        MOUSE_UP      : items.mvMouseReleaseHandler,
        MOUSE_MOVE    : items.mvMouseMoveHandler,
        MOUSE_DRAG    : items.mvMouseDragHandler,
        MOUSE_V_SCROLL: items.mvMouseWheelHandler,
    }

    _inputs: collections.deque[tuple[float, InputEvent, Any]]

    @overload
    def __init__(   # type: ignore
        self,
        maxlen   : int                    = 10000,
        events   : Collection[InputEvent] = InputEvent,
        inclusive: bool                   = True,
        *,
        label             : str  = ...,
        user_data         : Any  = ...,
        use_internal_label: bool = ...,
        tag               : Item = ...,
        show              : bool = ...,
        **kwargs
    ) -> None: ...
    def __init__(self, maxlen: int = 10000, events: Collection[InputEvent] = InputEvent, inclusive: bool = True, **kwargs):
        """Args:
            * maxlen: The length of the cache. Will grow to an
            arbitrary length when set to 0 (note: not a good idea).

            * events: A collection of `InputEvent` members of events
            to capture or ignore, depending on the value of *inclusive*.

            * inclusive: If True, only capture inputs specified in
            *events*. If False, capture all inputs except those
            specified in *events*.
        """
        super().__init__(**kwargs)
        self._inputs = collections.deque(
            [(time.time(), InputEvent.NONE, 0)],
            maxlen=maxlen
        )
        with self:
            for input_tp, h_factory in self.__INPUT_HOOK_MAP.items():
                if inclusive:
                    if input_tp in events:
                        h_factory(callback=getattr(self, f'_cb_{input_tp.lower()}'))
                elif input_tp not in events:
                    h_factory(callback=getattr(self, f'_cb_{input_tp.lower()}'))

    @property
    def oldest(self):
        with self.lock:
            return self._inputs[0]

    @property
    def newest(self):
        with self.lock:
            return self._inputs[-1]


class KeyPoller(_InputPoller):
    """A global handler registry that polls and caches keycode
    inputs.

    The registry is a living cache and behaves like a primitive
    sequence. Iterating through it yields the keys held down at
    the time of the interaction from oldest to newest. It also
    supports indexing and slicing.

    Values of the most recent mouse inputs are cached and
    are exposed via instance attributes;
        * last_pressed: The event code of the most recently-pressed
        key.

        * last_released: The event code of the most recently-
        released key.

    "Push"/"click" events are captured as individual "down"
    and "up" events.
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        with self:
            items.mvKeyDownHandler(callback=self._cb_key_down)
            items.mvKeyReleaseHandler(callback=self._cb_key_up)

    def to_chars(self) -> str:
        with self.lock:
            return ''.join(chr(i) for i in self._inputs)

    def keys_pressed(self, *keys: int) -> bool:
        """Return True if a series of keys are currently pressed
        and were pressed in order.
        """
        # BUG: This can return a false positive in niche situations.
        # For example; 'l_ctrl', 'l_shift', and 'f' are pressed in order
        # and are held down. Then, 'l_shift' is released -- passing a
        # key set of `(l_ctrl, f)` at this time would return True.
        with self.lock:
            return ' '.join(str(k) for k in keys) in ' '.join(str(i) for i in self._inputs)


    last_pressed = 0

    def _cb_key_down(self, handler: Item, key_info: tuple[int, float]):
        key = _KEYCODE_MAP[key_info[0]]
        with self.lock:
            self.last_pressed = key
            if key not in self._inputs:
                self._inputs.append(key)
            Py_DECREF(key_info)


    last_released = 0

    def _cb_key_up(self, handler: Item, key: int):
        key = _KEYCODE_MAP[key]
        with self.lock:
            self.last_released = key
            if key in self._inputs:
                self._inputs.remove(key)


class MousePoller(_InputPoller):
    """A global handler registry that polls and caches mouse
    inputs.

    The registry is a living cache and behaves like a primitive
    sequence. Iterating through it yields the mouse buttons held
    down at the time of the interaction from oldest to newest.
    It also supports indexing and slicing.

    Values of the most recent mouse inputs are cached and
    are exposed via instance attributes;
        * cursor_pos: A 2-item sequence of the last known coordinates
        of the mouse cursor, relative to the top-left corner of
        the Dear PyGui viewport.

        * last_pressed: The event code of the most recently-pressed
        mouse button.

        * last_released: The event code of the most recently-
        released mouse button.

        * last_scroll: A single integer indicating the direction
        and intensity of the most recent scroll input. The value
        is positive when a "scroll up" input is captured, and vice-
        versa. The speed or intensity of the input is indicated by
        the integer's absolute value; the higher the value, the
        greater the intensity.

        * last_drag_delta: A 3-item tuple containing the event
        code of the button held down and the x/y position delta
        of the cursor. Coordinates are relative to the location
        of the cursor when the button was initially pressed.

    "Push"/"click" events are captured as individual "down"
    and "up" events.
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        with self:
            items.mvMouseDownHandler(callback=self._cb_mouse_down)
            items.mvMouseReleaseHandler(callback=self._cb_mouse_up)
            items.mvMouseWheelHandler(callback=self._cb_mouse_scroll)
            items.mvMouseMoveHandler(callback=self._cb_mouse_move)
            items.mvMouseDragHandler(callback=self._cb_mouse_drag)


    last_pressed = 0

    def _cb_mouse_down(self, handler: Item, key_info: tuple[int, float]):
        mkey = _PTRCODE_MAP[key_info[0]]
        with self.lock:
            self.last_pressed = mkey
            if mkey not in self._inputs:
                self._inputs.append(mkey)
            Py_DECREF(key_info)


    last_released = 0

    def _cb_mouse_up(self, handler: Item, mkey: int):
        mkey = _PTRCODE_MAP[mkey]
        with self.lock:
            self.last_released = 0
            if mkey in self._inputs:
                self._inputs.remove(mkey)


    last_scroll = 0

    def _cb_mouse_scroll(self, handler: Item, mscroll: int):
        self.last_scroll = mscroll
        Py_DECREF(mscroll)


    last_drag_delta = (0, 0.0, 0.0)

    def _cb_mouse_drag(self, handler: Item, mdrag_info: tuple[int, float, float]):
        mkey, dt_xpos, dt_ypos =  mdrag_info
        self.last_drag_delta = _PTRCODE_MAP[mkey], dt_xpos, dt_ypos
        Py_DECREF(mdrag_info)


    cursor_pos = (0.0, 0.0)

    def _cb_mouse_move(self, handler: Item, pos: tuple[float, float]):
        self.cursor_pos = pos
        Py_DECREF(pos)









# [ `Callback`, `CallStack` ]

_CALLBACK_MARKER = '__item_callback__'


@_tools.frozen_namespace
class _CallArgs:
    SENDER    = 'sender'
    APP_DATA  = 'app_data'
    USER_DATA = 'user_data'


def _count_pargs(_callable: Callable):
    try:
        sig = inspect.signature(_callable)
    except TypeError:
        raise TypeError(f"{_callable!r} is not callable.") from None
    except ValueError: # "built-in"/ c-extension function?
        if _tools.is_builtin(_callable):
            raise ValueError(
                f"built-in object {_callable.__name__!r} has no signature."
            ) from None
        raise

    variadic_pos     = _Parameter.VAR_POSITIONAL
    positional_types = _Parameter.POSITIONAL_OR_KEYWORD, _Parameter.POSITIONAL_ONLY
    # It is not ideal to go through `__code__.co_argcount` as it won't include
    # variadic positional arguments.
    pos_arg_cnt = 0
    for p in sig.parameters.values():
        kind = p.kind
        if kind in positional_types:
            pos_arg_cnt += 1
        elif kind == variadic_pos:
            pos_arg_cnt = 3
            break
        # DearPyGui will only send a max of three positional args.
        if pos_arg_cnt >= 3:
            pos_arg_cnt = 3
            break
    return pos_arg_cnt




class Callback(ItemInterface):
    """Dynamic "Dear PyGui-callable" object wrapper. Enables non-
    function callables to be callable by Dear PyGui, allowing them
    to be registered as callbacks. In addition, optionally set
    default values for the callback's positional arguments which
    will override Dear PyGui-sent positional arguments.

    `Callback` instances can be compared to a less-generalized
    `functools.partial`. While they are specifically used for
    making a callable "Dear PyGui-callable", they are fairly multi-
    purpose and can be used in place of many lambdas, partials, closures,
    or decorators. A rather niche use-case for them is in priority
    queues, as wrappers support "priority" rich comparisons. Unlike
    `functools.partial` objects, argument overrides do not affect the
    the wrapper call procedure or signature -- "sender", "app_data",
    and "user_data" positional arguments are always accepted (but will
    be ignored when using an override). Overridden arguments can also
    be updated throughout the wrapper's lifetime.

    The implementation of this structure is aimed at keeping the on-call
    overhead onset by the wrapper to an minimum. The `__call__` method
    only forwards calls the target callback and does nothing else; no
    logic, no comparisons - no other code is executed. Instead, the
    overhead is "front-loaded" onto the first call to the wrapper
    proceeding an update to the assigned callback or overridding
    arguments; compiling a new `__call__` method using the current
    state of the wrapper. Updating the wrapper's state will force the
    creation of a new `__call__` method on the next call to the wrapper.
    This can be done beforehand by calling the `.prepare` method.

    The wrapper holds a non-reentrant lock for all reads and writes
    to help maintain thread safety. The lock is not held when the
    wrapper is called.

    The wrapper does not modify the target callback object. They remain
    as they were when set on the wrapper, and can be accessed or updated
    at any time via the `.callback` attribute.
    """
    __slots__ = (
        '__call__',
        '__code__',
        '__wrapped__',
        '_call_args',
        '_lock',
        '_priority',
    )

    __item_callback__ = True

    empty = _empty

    priority: Property[int] = _tools.simpleproperty()

    def __init__(
        self,
        callback : Callable | None = None,
        /,
        sender   : Item | Empty = empty,
        app_data : Any  | Empty = empty,
        user_data: Any  | Empty = empty,
        *,
        priority : int         = 1,
    ) -> None:
        """Args:
            * callback: The target callable to wrap.

            * sender: Override for the callback's first positional argument
            "sender".

            * app_data: Override for the callback's second positional
            argument "app_data".

            * user_data: Override for the callback's third positional argument
            "user_data".

            * priority: Numeric value used for rich comparisons.

        The *priority* value is used when comparing the wrapper to other
        objects; useful when used with priority queues.

        """
        self.priority    = priority
        self.__call__    = self.__call
        self.__code__    = self.__call.__code__
        self.__wrapped__ = None
        self._call_args  = (sender, app_data, user_data)
        self._lock       = threading.RLock()
        self.configure(
            callback=callback,
            sender=sender,
            app_data=app_data,
            user_data=user_data
        )

    def __lt__(self, other: Any):
        return self.priority < other

    def __le__(self, other: Any):
        return self.priority <= other

    def __eq__(self, other: Any):
        return self.priority == other

    def __ge__(self, other: Any):
        return self.priority >= other

    def __gt__(self, other: Any):
        return self.priority > other

    def __call__(self, sender: Item = 0, app_data: Any = None, user_data: Any = None, /) -> None: ...
    exec('del __call__')  # signature for auto-complete -- hide removal from pyright

    callback : Property[Callable | None] = ItemConfig()
    sender   : Property[Item | Empty]    = ItemConfig()
    app_data : Property[Any | Empty]     = ItemConfig()
    user_data: Property[Any | Empty]     = ItemConfig()

    @property
    def args(self):
        return self._call_args
    @args.setter
    def args(self, value: Array[Item | Empty, Any | Empty, Any | Empty] | Empty | None):
        try:
            sender, app_data, user_data = value  # type:ignore
        except TypeError:
            if value in (None, _empty):
                sender = app_data = user_data = _empty
            raise
        self.configure(
            sender=sender,
            app_data=app_data,
            user_data=user_data
        )

    @overload
    @override
    def configure(  # type:ignore
        self,
        *,
        callback : Callable | None = ...,
        sender   : Item | Empty    = ...,
        app_data : Any | Empty     = ...,
        user_data: Any | Empty     = ...,
    ) -> None: ...
    @override
    def configure(self, **kwargs):
        with self._lock:
            callback, *call_args = (
                kwargs.pop(k, v)
                for k, v in zip(
                    ('callback', _CallArgs.SENDER, _CallArgs.APP_DATA, _CallArgs.USER_DATA),
                    (self.__wrapped__, *self._call_args)
                )
            )
            if kwargs:
                raise TypeError(
                    f'{self.configure.__name__}() got unexpected keyword '
                    f'argument(s) {", ".join(repr(k) for k in kwargs)}.'
                )
            self.__call__    = self.__call
            self.__code__    = self.__call.__code__
            self.__wrapped__ = callback
            self._call_args  = tuple(call_args)

    @override
    def configuration(self):
        with self._lock:
            return dict(zip(
                ('callback', _CallArgs.SENDER, _CallArgs.APP_DATA, _CallArgs.USER_DATA),
                (self.__wrapped__, *self._call_args)),
            )

    @override
    def information(self):
        return {'priority': self.priority}

    __CALL_ARGS = (
        'self',
        f'{_CallArgs.SENDER}: Item = 0',
        f'{_CallArgs.APP_DATA}: Any = None',
        f'{_CallArgs.USER_DATA}: Any = None',
        # BUG: `__call__` must be able to accept a total of 5 args, including
        # `self`. This is because DPG sees `self` in `__code__.co_argcount` and
        # tries to pass an additional argument. This wouldn't be an issue of it
        # called `__call__` unbound, but the logic expects functions, so...
        '*args'
    )
    __CALL_LCLS  = (
        f'_{_CallArgs.SENDER}_OVERRIDE',
        f'_{_CallArgs.APP_DATA}_OVERRIDE',
        f'_{_CallArgs.USER_DATA}_OVERRIDE',
    )
    __CALLBACK_ARGS = tuple(zip(
        (_CallArgs.SENDER, _CallArgs.APP_DATA, _CallArgs.USER_DATA),
        (__CALL_LCLS)
    ))

    def __call(self, *args) -> None:
        if self.callback is not None:
            self.prepare()
            self.__call__(*args)

    def prepare(self):
        """Write, compile, and set this object's `__call__` method, binding
        any positional argument overrides.

        This is done automatically when the object is first called after
        updating the "callback" or positional argument override attributes.
        It can, however, be called manually in advance.
        """
        with self._lock:
            callback = self.__wrapped__
            try:
                arg_count = _count_pargs(callback)  # type:ignore
            except TypeError:
                raise TypeError(f"`callback` value must be callable.")
            except ValueError:
                raise ValueError(f"unable to get `callback` value signature")

            # mimic what DPG does and only send args if the callback
            # can accept them
            call_locals = dict(zip(self.__CALL_LCLS, self._call_args))
            call_locals['_callback'] = callback
            callback_args = (
                passed_arg if call_locals[ovrrd_arg] is _empty else ovrrd_arg
                for passed_arg, ovrrd_arg in self.__CALLBACK_ARGS[:arg_count]
            )
            __call__ = _tools.create_function(
                '__call__',
                self.__CALL_ARGS,
                (f'_callback({", ".join(callback_args)})',),
                None,
                globals=globals(),
                locals=call_locals,
            )
            self.__call__ = MethodType(__call__, self)
            self.__code__ = self.__call__.__code__  # type:ignore


class CallStack(Callback):
    """A list-like Dear PyGui-callable object containing other
    callables. When called, callbacks are executed in FIFO order
    without clearing them from the queue. Like `Callback` wrapper
    objects, stack-level positional argument overrides are supported.

    Every stack operation is thread-safe. Callbacks executed by
    the stack can even edit the call stack's contents. However,
    note that the stack takes a "snapshot" of its' current state
    and workload before processing, so changes made to the stack
    at that time won't be reflected on the current workload.

    Callbacks are stored in a `list` object and are exposed via the
    read-only `.callbacks` attribute. Operations performed directly
    on this object are NOT thread-safe. For this reason, it is
    recommended to operate on the stack instance instead. If you choose
    to operate on the underlying sequence instead, it is best to
    use the stack as a context manager as it will automatically manage
    the internal lock and return the underlying callback container.
        >>> stack = Callstack()
        >>> with stack as callbacks:
        ...     ...

    When called, the stack will try to pass three positional
    arguments to each callback; `sender`, `app_data` `user_data`.
    Each callable added to the stack is first checked for
    compatibility; it will be wrapped in a `Callback` object
    beforehand if it cannot accept all three arguments. Both the
    compatibility check and the wrapping are costly procedures.
    If you are certain that your callables adhere to the protocol,
    you can operate on the internal callback container directly with
    the procedure mentioned above.

    The `.append` and `.appendleft` methods return the (wrapped)
    object, allowing them to be used as decorators.

    Since the stack will wrap any callable that doesn't fit its'
    "schema", they can technically contain any callable object. This
    includes other other stacks. However, adding a stack to itself
    via `.append` or `.appendleft` will likely result in a
    `RecursionError`.

    `Callstack` inherits `Callback`s rich comparison behavior. This
    can be misleading since the behavior differs from that of other
    sequence types.

    The underlying lock IS reentrant, unlike the lock used by
    `Callback` objects.
    """
    __slots__ = ("_callbacks",)

    def __init_subclass__(cls):
        super().__init_subclass__()
        # ensure the 'DPG-callable hack' persists when overriding `__call__`
        if '__call__' in cls.__dict__:
            cls.__code__ = cls.__call__.__code__

    def __init__(
        self,
        iterable : Iterable[Callable] = (),
        /,
        sender   : Item | Empty = _empty,
        app_data : Any  | Empty = _empty,
        user_data: Any  | Empty = _empty,
        *,
        priority : int         = 1,
    ) -> None:
        """Args:
            * iterable: Inital contents for the stack.

            * sender: Override for the callback's first positional argument
            "sender".

            * app_data: Override for the callback's second positional
            argument "app_data".

            * user_data: Override for the callback's third positional argument
            "user_data".

            * priority: Numeric value used for rich comparisons.

        The *priority* value is used when comparing the wrapper to other
        objects; useful when used with priority queues.
        """
        # a reentrant lock is used here due to the heightened chance
        # of deadlocking -- mainly due to the object being iterable
        self._lock       = threading.RLock()
        self._call_args  = (sender, app_data, user_data)
        self.priority    = priority
        self.extend(iterable)

    # BUG: `__call__` must be able to accept a total of 5
    # args, including `self`. This is because DPG sees `self`
    # in `__code__.co_argcount` and tries to pass an additional
    # argument. This wouldn't be an issue of it called `__call__`
    # unbound, but the logic expects functions, so...
    @overload
    def __call__(self, sender: Item = 0, app_data: Any = None, user_data: Any = None, /) -> None: ...  # type:ignore
    def __call__(self, sender: Item = 0, app_data: Any = None, user_data: Any = None, *args) -> None:
        # The lock is not held during execution so callbacks can
        # edit the stack without issue.
        with self._lock:
            # Creating a shallow-copy of the queue is the easiest
            # way to make this op thread-safe while solving key
            # problems. Not the most performant, though...
            callbacks = self.callbacks.copy()
            sender, app_data, user_data = [  # type: ignore
                passed_arg if ovrrd_arg is _empty else ovrrd_arg
                for passed_arg, ovrrd_arg in zip((sender, app_data, user_data), self._call_args)
            ]
        for callback in callbacks:
            callback(sender, app_data, user_data)

    __code__ = __call__.__code__

    def __enter__(self):
        self._lock.acquire()
        return self.callbacks

    def __exit__(self, *args):
        self._lock.release()

    callbacks: Property[list[Callable]] = _tools.simpleproperty()

    @property
    def callback(self):
        return self.__call__

    @overload
    @override
    def configure(  # type:ignore
        self,
        *,
        sender   : Item | _empty = ...,
        app_data : Any | _empty  = ...,
        user_data: Any | _empty  = ...,
    ) -> None: ...
    @override
    def configure(self, **kwargs):
        with self._lock:
            call_args = (
                kwargs.pop(k, v)
                for k, v in zip(
                    (_CallArgs.SENDER, _CallArgs.APP_DATA, _CallArgs.USER_DATA),
                    self._call_args
                )
            )
            if kwargs:
                raise TypeError(
                    f'{self.configure.__name__}() got unexpected keyword '
                    f'argument(s) {", ".join(repr(k) for k in kwargs)}.'
                )
            self._call_args  = tuple(call_args)

    @override
    def configuration(self):
        with self._lock:
            return dict(zip(
                (_CallArgs.SENDER, _CallArgs.APP_DATA, _CallArgs.USER_DATA),
                self._call_args),
            )

    @override
    def prepare(self):
        """Not implemented."""

    # [ INTERNAL HELPERS ]

    def _prepped_callback(self, value: Callable):
        if (
            not hasattr(value, _CALLBACK_MARKER)
            or _count_pargs(value) < 3
        ):
            return Callback(value)
        return value

    def _prepped_callbacks(self, values: Iterable[Callable]):
        if hasattr(values, _CALLBACK_MARKER):
            return values
        yield from (Callback(v) for v in values)

    # [ QUEUE METHODS ]

    def __bool__(self) -> bool:
        with self._lock:
            return len(self.callbacks) == 0

    def __len__(self) -> int:
        with self._lock:
            return len(self.callbacks)

    def __iter__(self):
        with self._lock:
            yield from self.callbacks

    def __reversed__(self):
        with self._lock:
            yield from self.callbacks.__reversed__()

    def __contains__(self, value: Any):
        with self._lock:
            return value in self.callbacks

    def __getitem__(self, index: SupportsIndex):
        with self._lock:
            return self.callbacks[index]

    def __setitem__(self, index: SupportsIndex, value: Callable):
        with self._lock:
            self.callbacks[index] = self._prepped_callback(value)

    def __delitem__(self, index: SupportsIndex):
        with self._lock:
            del self.callbacks[index]

    def count(self, value: Any):
        with self._lock:
            return self.callbacks.count(value)

    def insert(self, index: SupportsIndex, value: Callable):
        value = self._prepped_callback(value)
        with self._lock:
            self.callbacks.insert(index, value)
        return value  # decorator usage

    def append(self, value: Callable):
        """Add a callback to the end of the stack. Can be used as a decorator."""
        value = self._prepped_callback(value)
        with self._lock:
            self.callbacks.append(value)
        return value  # decorator usage

    def appendleft(self, value: Callable):
        """Add a callback to the start of the stack. Can be used as a decorator."""
        return self.insert(0, value)  # decorator usage

    def extend(self, value: Iterable[Callable]):
        with self._lock:
            value = list(self.callbacks) if value is self else self._prepped_callbacks(value)
            self.callbacks.extend(value)

    def __add__(self, value: Iterable[Callable]):
        self.extend(value)
        return self

    __iadd__ = __add__

    def extendleft(self, value: Iterable[Callable]):
        with self._lock:
            value = list(self.callbacks) if value is self else self._prepped_callbacks(value)
            callbacks = (*value, *self.callbacks)
            self.callbacks.clear()
            self.callbacks.extend(callbacks)

    def pop(self, index: SupportsIndex = -1):
        with self._lock:
            return self.callbacks.pop(index)

    def popleft(self):
        return self.pop(0)

    def clear(self):
        with self._lock:
            self.callbacks.clear()

    def copy(self) -> Self:
        copy = type(self)(
            (),
            priority=self.priority,
            **self.configuration()  # holds lock
        )
        with self._lock:
            copy.callbacks.extend(self.callbacks)   # avoid re-processing values
        return copy

    __copy__ = copy

    def reverse(self):
        with self._lock:
            self.callbacks.reverse()


Callstack = CallStack








# [ UTILITIES ]

@overload
def callback(*, sender: Item = ..., app_data: Any = ..., user_data: Any = ..., priority: int = ...) -> Callback: ...
@overload
def callback(callback: ItemCallback, /, sender: Item = ..., app_data: Any = ..., user_data: Any = ..., priority: int = ...) -> Callback: ...
def callback(callback: Any = None, sender: Any = _empty, app_data: Any = _empty, user_data: Any = _empty, priority: int = 1) -> Any:
    """Convenience decorator for wrapping a callable in a
    `Callback` object.

    See the `Callback` class for more information.
    """
    def capture_callback(callback: Callable | None):
        return Callback(
            callback,
            priority=priority,
            sender=sender,
            app_data=app_data,
            user_data=user_data
        )

    if callback is None:
        return capture_callback
    return capture_callback(callback)


@overload
def cooldown(timer_fn: Callable[[], _N], interval: _N, /, *, no_args: bool = False) -> Callable[[Callable[_P, Any]], Callable[_P, None]]: ...
@overload
def cooldown(timer_fn: Callable[[], _N], interval: _N, /, *, no_args: bool = True) -> Callable[[Callable[[], Any]], Callable[[], None]]: ...
@overload
def cooldown(timer_fn: Callable[[], _N], interval: _N, callback: Callable[_P, Any], /, *, no_args: bool = False) -> Callable[_P, None]: ...
@overload
def cooldown(timer_fn: Callable[[], _N], interval: _N, callback: Callable[_P, Any], /, *, no_args: bool = True) -> Callable[[], None]: ...
def cooldown(timer_fn: Callable[[], _N], interval: _N, callback: Any = None, /, *, no_args: bool = False):
    """Return a wrapper that, when called, executes the target
    callable only if enough time has elapsed since the last time
    it was ran. This limits how often it can be executed over a
    period of time. Calls made to the returned wrapper while the
    target function is "cooling down" are no-op's.

    Can be used as a function decorator.

    Args:
        * timer_fn: A zero-argument callable that returns a time.

        * interval: The cooldown timer. The unit of time should match
        that of *timer_fn*s returned value.

        * callback: The target callable to apply the cooldown to.
        If *no_args* is set, it must be callable without arguments.

        * no_args: Improves the efficieny of the returned function,
        but limits *callback* options to zero-argument callables.

    The time units of *interval* and the return of *timer_fn*
    should match. For example, if *timer_fn* returns a number of
    fractional seconds (i.e. `time.perf_counter`), then *interval*
    is also expected to be a number of seconds.

    By default, the target callback's metadata is passed onto
    the returned wrapper via `functools.wraps`. When the *no_args*
    flag is set, the `.__next__` method of the underlying generator
    object is returned instead of a wrapper function. In this case,
    metadata is not copied over.

    This function does not create or utilize threads.
    """
    def capture_callback(callback: Callable) -> Callable:
        if no_args:

            def recurring_tasker():  # type: ignore
                ts_last_update = timer_fn() - interval
                while True:
                    ts_this_update = timer_fn()
                    if ts_this_update - ts_last_update >= interval:
                        ts_last_update = ts_this_update
                        callback()
                    yield

            dispatcher = recurring_tasker().__next__  # type: ignore

        else:

            def recurring_tasker():
                ts_last_update = timer_fn() - interval
                args, kwds = yield
                while True:
                    ts_this_update = timer_fn()
                    if ts_this_update - ts_last_update >= interval:
                        ts_last_update = ts_this_update
                        callback(*args, **kwds)
                    args, kwds = yield

            @functools.wraps(callback)
            def dispatcher(*args, **kwargs) -> None:
                _dispatcher.send((args, kwargs))

            _dispatcher = recurring_tasker()
            next(_dispatcher)  # prime generator

        return dispatcher

    if callback is None:
        return capture_callback
    return capture_callback(callback)


@overload
def cooldown_s(interval: _N, /, *, no_args: bool = False) -> Callable[[Callable[_P, Any]], Callable[_P, None]]: ...
@overload
def cooldown_s(interval: _N, /, *, no_args: bool = True) -> Callable[[Callable[[], Any]], Callable[[], None]]: ...
@overload
def cooldown_s(interval: _N, callback: Callable[_P, Any], /, *, no_args: bool = False) -> Callable[_P, None]: ...
@overload
def cooldown_s(interval: _N, callback: Callable[_P, Any], /, *, no_args: bool = True) -> Callable[[], None]: ...
def cooldown_s(interval: _N, callback: Any = None, /, no_args: bool = False):  # type: ignore
    """Adds a cooldown to a callable with an interval in
    fractional seconds.

    Refer to the `cooldown` function for more information.
    """
    return cooldown(time.perf_counter, interval, callback, no_args=no_args)


def root_ihandler_registry(item: Item) -> items.mvItemHandlerRegistry:
    """Return the item handler registry bound to an item's
    root parent. If unbound, create a new registry,
    bind it to the root parent, and return the new registry.
    """
    root_parent = api.Item.root_parent(item)
    root_ihr = api.Item.information(root_parent)['handlers']
    if not root_ihr:
        root_ihr = items.mvItemHandlerRegistry()
        api.Item.set_handlers(root_parent, root_ihr)
        return root_ihr
    return items.mvItemHandlerRegistry.new(root_ihr)


def resized_fill_content_region(
    item: Item,
    *,
    autosize_x : bool = True,
    autosize_y : bool = True,
) -> items.ResizeHandler:
    """Emulate the behavior of assigning a "stretch" policy
    onto the target item; when its' root parent is resized,
    the target will be sized to consume the available space
    within its' direct parent.

    Args:
        * item: Target item to resize.

        * autosize_x: If True, the target item's width will be
        adjusted to fill its' parent's available space.

        * autosize_y: If True, the target item's height will be
        adjusted to fill its' parent's available space.


    A `TypeError` or `ValueError` is raised when one or more of
    the following conditions are not met prior to calling this
    function;
        - the target item must be non-root item

        - the target item must be sizable via 'width' and/or
        'height' configuration

        - the target item's direct parent must support the
        'content_region_avail' state

        - the target item's root parent must support the
        'resized' state

    This function attaches a `mvResizeHandler` onto the
    target item's root parent. If the root parent does not
    have a bound item handler registry, one is created and
    assigned prior to creating and attaching the handler.
    The created `mvResizeHandler` item is destroyed when the
    callback is triggered and either the target item or its'
    direct parent are destroyed.

    The target's root parent will almost always be a
    `mvWindowAppItem` item, which supports 'resized'. Parent
    items that support 'content_region_avail' include
    `mvChildWindow` and `mvGroup` items. The target can be
    any sizable item, such as a `mvButton`, `mvChildWindow`,
    or `mvPlot` item.
    """
    item        = _interface.mvAll(item)
    item_config = item.configuration()
    item_info   = item.information()
    if (
        (autosize_x and "width" not in item_config) or
        (autosize_y and "height" not in item_config)
    ):
        raise TypeError(
            f"{item_info['type'].split('::')[1]!r} item {item!r} cannot be sized."
        )

    parent = item_info['parent']
    if not parent:
        raise TypeError(
            f"target cannot be a top-level root item."
        )
    parent = _interface.mvAll(parent)

    if not parent.state()['content_region_avail']:
        _parent_tp = parent.information()['type'].split('::')[1]
        raise TypeError(
            f"parent {_parent_tp!r} item does not have a content region."
        )

    get_x_scr_max = dearpygui.get_x_scroll_max
    get_y_scr_max = dearpygui.get_y_scroll_max
    try:
        parent.get_x_scroll_pos()
    except:
        get_x_scr_max = get_y_scr_max = lambda x: 0

    root_parent = cast(_interface.mvAll, item.root_parent)
    root_info   = root_parent.information()
    if not root_info['resized_handler_applicable']:
        raise TypeError(
            f"top-level parent {root_info['type'].split('::')[1]!r} item "
            f"does not support resize handlers."
        )

    root_ihr = root_info['handlers']
    if not root_ihr:
        root_ihr = ItemHandlerRegistry()
        api.Item.set_handlers(root_parent, root_ihr)

    if autosize_x and autosize_y:
        cb_body = (
            "avail_wt, avail_ht = parent.state()['content_region_avail']",
            "item.configure(",
            "    width=avail_wt,",
            "    height=avail_ht,",
            ")",
            "split_frame()",
            "avail_wt, avail_ht = parent.state()['content_region_avail']",
            "item.configure(",
            "    width=max(1, avail_wt - get_x_scr_max(parent)),",
            "    height=max(1, avail_ht - get_y_scr_max(parent)),",
            ")",
        )
    elif autosize_x:
        cb_body = (
            "item.configure(width=parent.state()['content_region_avail'][0])",
            "split_frame()",
            "item.configure(",
            "    width=max(1, parent.state()['content_region_avail'][0] - get_x_scr_max(parent))",
            ")",
        )
    elif autosize_y:
        cb_body = (
            "item.configure(height=parent.state()['content_region_avail'][1])",
            "split_frame()",
            "item.configure(",
            "    height=max(1, parent.state()['content_region_avail'][1] - get_y_scr_max(parent))",
            ")",
        )
    else:
        cb_body = ("...",)

    cb_body = (
        "try:",
        *(f'    {ln}' for ln in cb_body),
        "except SystemError:",
        "    if not item.exists() or not parent.exists():",
        "        handler.callback = None",
        "        handler.destroy()",
        "        del callback",
        "    else:",
        "        raise",
    )

    handler  = items.ResizeHandler.new()
    callback = _tools.create_function(
        'callback',
        (),
        cb_body,
        globals=globals(),
        locals={
            'item'            : _interface.mvAll(item),
            'parent'          : _interface.mvAll(parent),
            'handler'         : handler,
            'get_x_scr_max'   : get_x_scr_max,
            'get_y_scr_max'   : get_y_scr_max,
            'split_frame'     : api.Runtime.split_frame,
        }
    )
    handler.init(callback=callback, parent=root_ihr)
    return handler



import inspect
from . import _typing, _parsing, _tools, _mkstub
from ._typing import overload, Item, Color, Any, Array
from .items import mvThemeColor, ThemeColor, mvThemeStyle, ThemeStyle


def _init_module():
    overloads = _typing.get_overloads(_add_theme_element)

    for ovld in overloads:
        try:
            ovld_sig = ovld.__signature__
        except AttributeError:
            ovld_sig = inspect.signature(ovld)
        try:
            kwargs_sig = _ITP_BASE.configure.__signature__
        except AttributeError:
            kwargs_sig = _ITP_BASE.configure.__signature__ = inspect.signature(
                _ITP_BASE.configure
        )

        ovld_args = {
            n:p for n, p in ovld_sig.parameters.items()
            if p.kind is p.POSITIONAL_ONLY
        }
        ovld_kwds = {
            n:p for n, p in kwargs_sig.parameters.items()
            if n not in ovld_args
            and p.kind in (p.POSITIONAL_OR_KEYWORD, p.KEYWORD_ONLY)}
        ovld_kwds.pop('self', None)
        ovld_kwds.pop('cls', None)
        ovld.__signature__ = ovld_sig.replace(
            parameters=list((ovld_args | ovld_kwds).values())
        )

    exported = _mkstub.Exported(__name__)

    ns = {}
    for elem_def in _parsing.color_definitions().values():
        fn = _tools.create_function(
            elem_def.name2, _DFN_ARGS,
            _DFN_BODY,
            _ITP_BASE,
            globals=globals(),
            locals={
                'itp'      : _ITP_BASE,
                'category' : int(elem_def.category),
                'target'   : elem_def.target,
            }
        )
        fn.__module__ = __name__

        # HACK: force register existing overloads
        for ovld in overloads:
            ovld.__name__     = fn.__name__
            ovld.__qualname__ = fn.__qualname__
            overload(ovld)

        exported[fn.__name__] = fn
        exported.aliased[elem_def.name1] = fn
        ns[elem_def.name2] = ns[elem_def.name1] = fn

    globals().update(ns)




_ITP_BASE = mvThemeColor
_DFN_ARGS = ('value: Any, /', '*args: float', '**kwargs')
_DFN_BODY = (
    "kwargs.update(",
    "    value=value if not args else (value, *args),",
    "    target=target, category=category,",
    ")",
    "return itp(**kwargs)"
)

@overload
def _add_theme_element(value: Color, /, **kwargs) -> _ITP_BASE: ...
@overload
def _add_theme_element(r: float, g: float, b: float, a: float = ..., /, **kwargs) -> _ITP_BASE: ...
def _add_theme_element(*args, **kwargs) -> _ITP_BASE: ...

_init_module()




def _fill_namespace():
    from . import _interface, _parsing, _mkstub

    exported = _mkstub.Exported(__name__)

    namespace = dict(
        exp for exp in _mkstub.Exported.fetch(_interface).items()
    )
    for tp_def in _parsing.item_definitions().values():
        if not tp_def:
            continue

        itp = exported(_interface.create_itemtype(tp_def))
        itp.__module__ = __name__
        namespace[itp.__qualname__] = itp
        # set an alias
        if itp.__qualname__ == "mvWindowAppItem":
            namespace['mvWindow'] = namespace['Window'] = itp
            exported(itp, 'mvWindow')
            exported(itp, 'Window')
        elif itp.__qualname__  == 'mvAnnotation':
            namespace['PlotAnnotation'] = itp
            exported(itp, 'PlotAnnotation')
        else:
            alias = itp.__qualname__.removeprefix('mv')
            if alias[0].isdigit():
                alias = f'{alias[2:]}{alias[0]}{alias[1].upper()}'
            assert alias not in namespace
            namespace[alias] = itp
            exported(itp, alias)

    globals().update(namespace)


_fill_namespace()
